from flyteidl.admin import matchable_resource_pb2 as _matchable_resource

from flytekit.models import common as _common


class MatchableResource(object):
    TASK_RESOURCE = _matchable_resource.TASK_RESOURCE
    CLUSTER_RESOURCE = _matchable_resource.CLUSTER_RESOURCE
    EXECUTION_QUEUE = _matchable_resource.EXECUTION_QUEUE
    EXECUTION_CLUSTER_LABEL = _matchable_resource.EXECUTION_CLUSTER_LABEL
    QUALITY_OF_SERVICE_SPECIFICATION = _matchable_resource.QUALITY_OF_SERVICE_SPECIFICATION
    PLUGIN_OVERRIDE = _matchable_resource.PLUGIN_OVERRIDE

    @classmethod
    def enum_to_string(cls, val):
        """
        :param int val:
        :rtype: Text
        """
        if val == cls.TASK_RESOURCE:
            return "TASK_RESOURCE"
        elif val == cls.CLUSTER_RESOURCE:
            return "CLUSTER_RESOURCE"
        elif val == cls.EXECUTION_QUEUE:
            return "EXECUTION_QUEUE"
        elif val == cls.EXECUTION_CLUSTER_LABEL:
            return "EXECUTION_CLUSTER_LABEL"
        elif val == cls.QUALITY_OF_SERVICE_SPECIFICATION:
            return "QUALITY_OF_SERVICE_SPECIFICATION"
        else:
            return "<UNKNOWN>"

    @classmethod
    def string_to_enum(cls, val):
        """
        :param Text val:
        :rtype: int
        """
        if val == "TASK_RESOURCE":
            return cls.TASK_RESOURCE
        elif val == "CLUSTER_RESOURCE":
            return cls.CLUSTER_RESOURCE
        elif val == "EXECUTION_QUEUE":
            return cls.EXECUTION_QUEUE
        elif val == "EXECUTION_CLUSTER_LABEL":
            return cls.EXECUTION_CLUSTER_LABEL
        elif val == cls.QUALITY_OF_SERVICE_SPECIFICATION:
            return "QUALITY_OF_SERVICE_SPECIFICATION"
        else:
            return "<UNKNOWN>"


class ClusterResourceAttributes(_common.FlyteIdlEntity):
    def __init__(self, attributes):
        """
        Custom resource attributes which will be applied in cluster resource creation (e.g. quotas).
        Dict keys are the *case-sensitive* names of variables in templatized resource files.
        Dict values should be the custom values which get substituted during resource creation.

        :param dict[Text, Text] attributes: Applied in cluster resource creation (e.g. quotas).
        """
        self._attributes = attributes

    @property
    def attributes(self):
        """
        Custom resource attributes which will be applied in cluster resource management
        :rtype: dict[Text, Text]
        """
        return self._attributes

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.matchable_resource_pb2.ClusterResourceAttributes
        """
        return _matchable_resource.ClusterResourceAttributes(
            attributes=self.attributes,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.matchable_resource_pb2.ClusterResourceAttributes pb2_object:
        :rtype: ClusterResourceAttributes
        """
        return cls(
            attributes=pb2_object.attributes,
        )


class ExecutionQueueAttributes(_common.FlyteIdlEntity):
    def __init__(self, tags):
        """
        Tags used for assigning execution queues for tasks matching a project, domain and optionally, workflow.

        :param list[Text] tags:
        """
        self._tags = tags

    @property
    def tags(self):
        """
        :rtype: list[Text]
        """
        return self._tags

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.matchable_resource_pb2.ExecutionQueueAttributes
        """
        return _matchable_resource.ExecutionQueueAttributes(
            tags=self.tags,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.matchable_resource_pb2.ExecutionQueueAttributes pb2_object:
        :rtype: ExecutionQueueAttributes
        """
        return cls(
            tags=pb2_object.tags,
        )


class ExecutionClusterLabel(_common.FlyteIdlEntity):
    def __init__(self, value):
        """
        Label value to determine where the execution will be run

        :param Text value:
        """
        self._value = value

    @property
    def value(self):
        """
        :rtype: Text
        """
        return self._value

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.matchable_resource_pb2.ExecutionClusterLabel
        """
        return _matchable_resource.ExecutionClusterLabel(
            value=self.value,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.matchable_resource_pb2.ExecutionClusterLabel pb2_object:
        :rtype: ExecutionClusterLabel
        """
        return cls(
            value=pb2_object.value,
        )


class PluginOverride(_common.FlyteIdlEntity):
    FAIL = _matchable_resource.PluginOverride.FAIL
    USE_DEFAULT = _matchable_resource.PluginOverride.USE_DEFAULT

    @classmethod
    def string_to_enum(cls, val):
        """
        :param Text val:
        :rtype: int
        """
        if val == "FAIL":
            return cls.FAIL
        elif val == "USE_DEFAULT":
            return cls.USE_DEFAULT
        else:
            return "<UNKNOWN>"

    def __init__(self, task_type, plugin_id, missing_plugin_behavior):
        """
        Alternate plugin implementations requested for a specific task type.

        :param Text task_type:
        :param: list[Text] plugin_id:
        :param int missing_plugin_behavior
        """
        self._task_type = task_type
        self._plugin_id = plugin_id
        self._missing_plugin_behavior = missing_plugin_behavior

    @property
    def task_type(self):
        """
        :rtype: Text
        """
        return self._task_type

    @property
    def plugin_id(self):
        """
        :rtype: list[Text]
        """
        return self._plugin_id

    @property
    def missing_plugin_behavior(self):
        """
        :rtype: int
        """
        return self._missing_plugin_behavior

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.matchable_resource_pb2.PluginOverride
        """
        return _matchable_resource.PluginOverride(
            task_type=self.task_type,
            plugin_id=self.plugin_id,
            missing_plugin_behavior=self.missing_plugin_behavior,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.matchable_resource_pb2.PluginOverride pb2_object:
        :rtype: PluginOverride
        """
        return cls(
            task_type=pb2_object.task_type,
            plugin_id=pb2_object.plugin_id,
            missing_plugin_behavior=pb2_object.missing_plugin_behavior,
        )


class PluginOverrides(_common.FlyteIdlEntity):
    def __init__(self, overrides):
        """
        Alternate plugin implementations for designated task types.

        :param list[PluginOverride] overrides:
        """
        self._overrides = overrides

    @property
    def overrides(self):
        """
        :rtype: list[PluginOverride]
        """
        return self._overrides

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.matchable_resource_pb2.PluginOverrides
        """
        return _matchable_resource.PluginOverrides(overrides=[override.to_flyte_idl() for override in self.overrides])

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.matchable_resource_pb2.PluginOverrides pb2_object:
        :rtype: PluginOverrides
        """
        return cls(overrides=[PluginOverride.from_flyte_idl(override) for override in pb2_object.overrides])


class MatchingAttributes(_common.FlyteIdlEntity):
    def __init__(
        self,
        cluster_resource_attributes=None,
        execution_queue_attributes=None,
        execution_cluster_label=None,
        plugin_overrides=None,
    ):
        """
        At most one target from cluster_resource_attributes, execution_queue_attributes or execution_cluster_label
            can be set.
        :param ClusterResourceAttributes cluster_resource_attributes:
        :param ExecutionQueueAttributes execution_queue_attributes:
        :param ExecutionClusterLabel execution_cluster_label:
        :param PluginOverrides plugin_overrides:
        """
        if cluster_resource_attributes:
            if execution_queue_attributes or execution_cluster_label or plugin_overrides:
                raise ValueError("Only one target can be set")
        elif execution_queue_attributes and (execution_cluster_label or plugin_overrides):
            raise ValueError("Only one target can be set")
        elif execution_cluster_label and plugin_overrides:
            raise ValueError("Only one target can be set")

        self._cluster_resource_attributes = cluster_resource_attributes
        self._execution_queue_attributes = execution_queue_attributes
        self._execution_cluster_label = execution_cluster_label
        self._plugin_overrides = plugin_overrides

    @property
    def cluster_resource_attributes(self):
        """
        Custom resource attributes which will be applied in cluster resource creation (e.g. quotas).
        :rtype: ClusterResourceAttributes
        """
        return self._cluster_resource_attributes

    @property
    def execution_queue_attributes(self):
        """
        Tags used for assigning execution queues for tasks.
        :rtype: ExecutionQueueAttributes
        """
        return self._execution_queue_attributes

    @property
    def execution_cluster_label(self):
        """
        Label value to determine where the execution will be run.
        :rtype: ExecutionClusterLabel
        """
        return self._execution_cluster_label

    @property
    def plugin_overrides(self):
        """
        Plugin implementation overrides for specific task types.
        :rtype: PluginOverrides
        """
        return self._plugin_overrides

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.matchable_resource_pb2.MatchingAttributes
        """
        return _matchable_resource.MatchingAttributes(
            cluster_resource_attributes=self.cluster_resource_attributes.to_flyte_idl()
            if self.cluster_resource_attributes
            else None,
            execution_queue_attributes=self.execution_queue_attributes.to_flyte_idl()
            if self.execution_queue_attributes
            else None,
            execution_cluster_label=self.execution_cluster_label.to_flyte_idl()
            if self.execution_cluster_label
            else None,
            plugin_overrides=self.plugin_overrides.to_flyte_idl() if self.plugin_overrides else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.matchable_resource_pb2.MatchingAttributes pb2_object:
        :rtype: MatchingAttributes
        """
        return cls(
            cluster_resource_attributes=ClusterResourceAttributes.from_flyte_idl(pb2_object.cluster_resource_attributes)
            if pb2_object.HasField("cluster_resource_attributes")
            else None,
            execution_queue_attributes=ExecutionQueueAttributes.from_flyte_idl(pb2_object.execution_queue_attributes)
            if pb2_object.HasField("execution_queue_attributes")
            else None,
            execution_cluster_label=ExecutionClusterLabel.from_flyte_idl(pb2_object.execution_cluster_label)
            if pb2_object.HasField("execution_cluster_label")
            else None,
            plugin_overrides=PluginOverrides.from_flyte_idl(pb2_object.plugin_overrides)
            if pb2_object.HasField("plugin_overrides")
            else None,
        )



import configparser
import typing

from flytekit.configuration import common as _config_common


def get_specified_images() -> typing.Dict[str, str]:
    """
    This section should contain options, where the option name is the friendly name of the image and the corresponding
    value is actual FQN of the image. Example of how the section is structured
    [images]
    my_image1=docker.io/flyte:tag
    # Note that the tag is optional. If not specified it will be the default version identifier specified
    my_image2=docker.io/flyte

    :returns a dictionary of name: image<fqn+version> Version is optional
    """
    images: typing.Dict[str, str] = {}
    if _config_common.CONFIGURATION_SINGLETON.config is None:
        return images
    try:
        image_names = _config_common.CONFIGURATION_SINGLETON.config.options("images")
    except configparser.NoSectionError:
        print("No images specified, will use the default image")
        image_names = None
    if image_names:
        for i in image_names:
            images[str(i)] = _config_common.FlyteStringConfigurationEntry("images", i).get()
    return images



import abc as _abc
import datetime as _datetime
import time as _time

from flytekit.common.exceptions import user as _user_exceptions
from flytekit.models import common as _common_models


class ExecutionArtifact(object, metaclass=_common_models.FlyteABCMeta):
    @_abc.abstractproperty
    def inputs(self):
        """
        Returns the inputs to the execution in the standard Python format as dictated by the type engine.
        :rtype:  dict[Text, T]
        """
        pass

    @_abc.abstractproperty
    def outputs(self):
        """
        Returns the outputs to the execution in the standard Python format as dictated by the type engine.  If the
        execution ended in error or the execution is in progress, an exception will be raised.
        :rtype:  dict[Text, T]
        """
        pass

    @_abc.abstractproperty
    def error(self):
        """
        If execution is in progress, raise an exception.  Otherwise, return None if no error was present upon
        reaching completion.
        :rtype: flytekit.models.core.execution.ExecutionError or None
        """
        pass

    @_abc.abstractproperty
    def is_complete(self):
        """
        Dictates whether or not the execution is complete.
        :rtype: bool
        """
        pass

    @_abc.abstractmethod
    def sync(self):
        """
        Syncs the state of the underlying execution artifact with the state observed by the platform.
        :rtype: None
        """
        pass

    @_abc.abstractmethod
    def _sync_closure(self):
        """
        Syncs the closure of the underlying execution artifact with the state observed by the platform.
        :rtype: None
        """
        pass

    def wait_for_completion(self, timeout=None, poll_interval=None):
        """
        :param datetime.timedelta timeout: Amount of time to wait until the execution has completed before timing
            out. If not set or set to None, this method will wait for infinite.
        :param datetime.timedelta poll_interval: Duration to wait between polling for a completion update.
        :rtype: None
        """
        poll_interval = poll_interval or _datetime.timedelta(seconds=30)
        if timeout is None:
            time_to_give_up = _datetime.datetime.max
        else:
            time_to_give_up = _datetime.datetime.utcnow() + timeout

        self._sync_closure()
        while _datetime.datetime.utcnow() < time_to_give_up:
            if self.is_complete:
                self.sync()
                return
            _time.sleep(poll_interval.total_seconds())
            self._sync_closure()
        raise _user_exceptions.FlyteTimeout("Execution {} did not complete before timeout.".format(self))



from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import base_sdk_types as _base_sdk_types
from flytekit.common.types.impl import blobs as _blob_impl
from flytekit.models import literals as _literals
from flytekit.models import types as _idl_types
from flytekit.models.core import types as _core_types


class BlobInstantiator(_base_sdk_types.InstantiableType):
    @staticmethod
    def create_at_known_location(location):
        """
        :param Text location:
        :rtype: flytekit.common.types.impl.blobs.Blob
        """
        return _blob_impl.Blob.create_at_known_location(location, mode="wb")

    @staticmethod
    def fetch(remote_path, local_path=None):
        """
        :param Text remote_path:
        :param Text local_path: [Optional] If specified, the Blob is copied to this location.  If specified,
            this location is NOT managed and the blob will not be cleaned up upon exit.
        :rtype: flytekit.common.types.impl.blobs.Blob
        """
        return _blob_impl.Blob.fetch(remote_path, mode="rb", local_path=local_path)

    def __call__(cls, *args, **kwargs):
        """
        TODO: Is there a better way to deal with this?

        We want the behavior of Types.Blob() returns a _blob_impl.Blob, but also to be able to use this object to
        wrap a _blob_impl.Blob via Types.Blob(_blob_impl.Blob()) for serialization, type checking, etc..

        :rtype: flytekit.common.types.impl.blobs.Blob
        """
        if not args and not kwargs:
            return _blob_impl.Blob.create_at_any_location(mode="wb")
        else:
            return super(BlobInstantiator, cls).__call__(*args, **kwargs)


# TODO: Make blobs and schemas pluggable
class Blob(_base_sdk_types.FlyteSdkValue, metaclass=BlobInstantiator):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: Blob
        """
        if not string_value:
            _user_exceptions.FlyteValueException(string_value, "Cannot create a Blob from the provided path value.")
        return cls(_blob_impl.Blob.from_string(string_value, mode="rb"))

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif isinstance(t_value, _blob_impl.Blob):
            blob = t_value
        else:
            blob = _blob_impl.Blob.from_python_std(t_value)
        return cls(blob)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(
            blob=_core_types.BlobType(format="", dimensionality=_core_types.BlobType.BlobDimensionality.SINGLE)
        )

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Blob
        """
        return cls(_blob_impl.Blob.promote_from_model(literal_model.scalar.blob))

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Blob"

    def __init__(self, value):
        """
        :param flytekit.common.types.impl.blobs.Blob value: Blob value to wrap
        """
        super(Blob, self).__init__(scalar=_literals.Scalar(blob=value))

    def to_python_std(self):
        """
        :rtype: flytekit.common.types.impl.blobs.Blob
        """
        return self.scalar.blob

    def short_string(self):
        """
        :rtype: Text
        """
        return "Blob(uri={}{})".format(
            self.scalar.blob.uri,
            ", format={}".format(self.scalar.blob.metadata.type.format)
            if self.scalar.blob.metadata.type.format
            else "",
        )


class MultiPartBlobInstantiator(_base_sdk_types.InstantiableType):
    @staticmethod
    def create_at_known_location(location):
        """
        :param Text location:
        :rtype: flytekit.common.types.impl.blobs.MultiPartBlob
        """
        return _blob_impl.MultiPartBlob.create_at_known_location(location, mode="wb")

    @staticmethod
    def fetch(remote_path, local_path=None):
        """
        :param Text remote_path:
        :param Text local_path: [Optional] If specified, the MultiPartBlob is copied to this location.  If specified,
            this location is NOT managed and the blob will not be cleaned up upon exit.
        :rtype: flytekit.common.types.impl.blobs.MultiPartBlob
        """
        return _blob_impl.MultiPartBlob.fetch(remote_path, mode="rb", local_path=local_path)

    def __call__(cls, *args, **kwargs):
        """
        TODO: Is there a better way to deal with this?

        We want the behavior of Types.MultiPartBlob() returns a _blob_impl.MultiPartBlob, but also to be able to use
        this object to wrap a _blob_impl.MultiPartBlob via Types.MultiPartBlob(_blob_impl.MultiPartBlob()) for
        serialization, type checking, etc..

        :rtype: flytekit.common.types.impl.blobs.MultiPartBlob
        """
        if not args and not kwargs:
            return _blob_impl.MultiPartBlob.create_at_any_location(mode="wb")
        else:
            return super(MultiPartBlobInstantiator, cls).__call__(*args, **kwargs)


class MultiPartBlob(_base_sdk_types.FlyteSdkValue, metaclass=MultiPartBlobInstantiator):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: MultiPartBlob
        """
        if not string_value:
            _user_exceptions.FlyteValueException(
                string_value,
                "Cannot create a MultiPartBlob from the provided path " "value.",
            )
        return cls(_blob_impl.MultiPartBlob.from_string(string_value, mode="rb"))

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif isinstance(t_value, _blob_impl.MultiPartBlob):
            blob = t_value
        else:
            blob = _blob_impl.MultiPartBlob.from_python_std(t_value)
        return cls(blob)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(
            blob=_core_types.BlobType(
                format="",
                dimensionality=_core_types.BlobType.BlobDimensionality.MULTIPART,
            )
        )

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: MultiPartBlob
        """
        return cls(_blob_impl.MultiPartBlob.promote_from_model(literal_model.scalar.blob))

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "MultiPartBlob"

    def __init__(self, value):
        """
        :param flytekit.common.types.impl.blobs.MultiPartBlob value: Blob value to wrap
        """
        super(MultiPartBlob, self).__init__(scalar=_literals.Scalar(blob=value))

    def to_python_std(self):
        """
        :rtype: flytekit.common.types.impl.blobs.MultiPartBlob
        """
        return self.scalar.blob

    def short_string(self):
        """
        :rtype: Text
        """
        return "MultiPartBlob(uri={}{})".format(
            self.scalar.blob.uri,
            ", format={}".format(self.scalar.blob.metadata.type.format)
            if self.scalar.blob.metadata.type.format
            else "",
        )


class CsvInstantiator(BlobInstantiator):
    @staticmethod
    def create_at_known_location(location):
        """
        :param Text location:
        :rtype: flytekit.common.types.impl.blobs.CSV
        """
        return _blob_impl.Blob.create_at_known_location(location, mode="w", format="csv")

    @staticmethod
    def fetch(remote_path, local_path=None):
        """
        :param Text remote_path:
        :param Text local_path: [Optional] If specified, the MultiPartBlob is copied to this location.  If specified,
            this location is NOT managed and the blob will not be cleaned up upon exit.
        :rtype: flytekit.common.types.impl.blobs.CSV
        """
        return _blob_impl.Blob.fetch(remote_path, local_path=local_path, mode="r", format="csv")

    def __call__(cls, *args, **kwargs):
        """
        TODO: Is there a better way to deal with this?

        We want the behavior of Types.CSV() returns a _blob_impl.CSV, but also to be able to use
        this object to wrap a _blob_impl.CSV via Types.CSV(_blob_impl.CSV()) for
        serialization, type checking, etc..

        :rtype: flytekit.common.types.impl.blobs.CSV
        """
        if not args and not kwargs:
            return _blob_impl.Blob.create_at_any_location(mode="w", format="csv")
        else:
            return super(CsvInstantiator, cls).__call__(*args, **kwargs)


class CSV(Blob, metaclass=CsvInstantiator):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: CSV
        """
        if not string_value:
            _user_exceptions.FlyteValueException(string_value, "Cannot create a CSV from the provided path value.")
        return cls(_blob_impl.Blob.from_string(string_value, format="csv", mode="r"))

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif isinstance(t_value, _blob_impl.Blob):
            if t_value.metadata.type.format != "csv":
                raise _user_exceptions.FlyteValueException(t_value, "Blob is in incorrect format.  Expected CSV.")
            blob = t_value
        else:
            blob = _blob_impl.Blob.from_python_std(t_value, format="csv", mode="w")
        return cls(blob)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(
            blob=_core_types.BlobType(
                format="csv",
                dimensionality=_core_types.BlobType.BlobDimensionality.SINGLE,
            )
        )

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: CSV
        """
        return cls(_blob_impl.Blob.promote_from_model(literal_model.scalar.blob, mode="r"))

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "CSV"

    def __init__(self, value):
        """
        :param flytekit.common.types.impl.blobs.Blob value: CSV blob value to wrap
        """
        super(CSV, self).__init__(value)


class MultiPartCsvInstantiator(MultiPartBlobInstantiator):
    @staticmethod
    def create_at_known_location(location):
        """
        :param Text location:
        :rtype: flytekit.common.types.impl.blobs.MultiPartBlob
        """
        return _blob_impl.MultiPartBlob.create_at_known_location(location, mode="w", format="csv")

    @staticmethod
    def fetch(remote_path, local_path=None):
        """
        :param Text remote_path:
        :param Text local_path: [Optional] If specified, the MultiPartCSV is copied to this location.  If specified,
            this location is NOT managed and the blob will not be cleaned up upon exit.
        :rtype: flytekit.common.types.impl.blobs.MultiPartCSV
        """
        return _blob_impl.MultiPartBlob.fetch(remote_path, local_path=local_path, mode="r", format="csv")

    def __call__(cls, *args, **kwargs):
        """
        TODO: Is there a better way to deal with this?

        We want the behavior of Types.MultiPartCSV() returns a _blob_impl.MultiPartCSV, but also to be able to use
        this object to wrap a _blob_impl.MultiPartCSV via Types.MultiPartCSV(_blob_impl.MultiPartCSV()) for
        serialization, type checking, etc..

        :rtype: flytekit.common.types.impl.blobs.MultiPartCSV
        """
        if not args and not kwargs:
            return _blob_impl.MultiPartBlob.create_at_any_location(mode="w", format="csv")
        else:
            return super(MultiPartCsvInstantiator, cls).__call__(*args, **kwargs)


class MultiPartCSV(MultiPartBlob, metaclass=MultiPartCsvInstantiator):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: MultiPartCSV
        """
        if not string_value:
            _user_exceptions.FlyteValueException(
                string_value,
                "Cannot create a MultiPartCSV from the provided path value.",
            )
        return cls(_blob_impl.MultiPartBlob.from_string(string_value, format="csv", mode="r"))

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif isinstance(t_value, _blob_impl.MultiPartBlob):
            if t_value.metadata.type.format != "csv":
                raise _user_exceptions.FlyteValueException(
                    t_value, "Multi Part Blob is in incorrect format.  Expected CSV."
                )
            blob = t_value
        else:
            blob = _blob_impl.MultiPartBlob.from_python_std(t_value, format="csv", mode="w")
        return cls(blob)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(
            blob=_core_types.BlobType(
                format="csv",
                dimensionality=_core_types.BlobType.BlobDimensionality.MULTIPART,
            )
        )

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: MultiPartCSV
        """
        return cls(_blob_impl.MultiPartBlob.promote_from_model(literal_model.scalar.blob, mode="r"))

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "MultiPartCSV"

    def __init__(self, value):
        """
        :param flytekit.common.types.impl.blobs.MultiPartBlob value: MultiPartBlob value to wrap
        """
        super(MultiPartCSV, self).__init__(value)



import flyteidl.admin.node_execution_pb2 as _node_execution_pb2
import pytz as _pytz

from flytekit.models import common as _common_models
from flytekit.models.core import execution as _core_execution
from flytekit.models.core import identifier as _identifier


class NodeExecutionClosure(_common_models.FlyteIdlEntity):
    def __init__(self, phase, started_at, duration, output_uri=None, error=None):
        """
        :param int phase:
        :param datetime.datetime started_at:
        :param datetime.timedelta duration:
        :param Text output_uri:
        :param flytekit.models.core.execution.ExecutionError error:
        """
        self._phase = phase
        self._started_at = started_at
        self._duration = duration
        self._output_uri = output_uri
        self._error = error

    @property
    def phase(self):
        """
        :rtype: int
        """
        return self._phase

    @property
    def started_at(self):
        """
        :rtype: datetime.datetime
        """
        return self._started_at

    @property
    def duration(self):
        """
        :rtype: datetime.timedelta
        """
        return self._duration

    @property
    def output_uri(self):
        """
        :rtype: Text
        """
        return self._output_uri

    @property
    def error(self):
        """
        :rtype: flytekit.models.core.execution.ExecutionError
        """
        return self._error

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.node_execution_pb2.NodeExecutionClosure
        """
        obj = _node_execution_pb2.NodeExecutionClosure(
            phase=self.phase,
            output_uri=self.output_uri,
            error=self.error.to_flyte_idl() if self.error is not None else None,
        )
        obj.started_at.FromDatetime(self.started_at.astimezone(_pytz.UTC).replace(tzinfo=None))
        obj.duration.FromTimedelta(self.duration)
        return obj

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.admin.node_execution_pb2.NodeExecutionClosure p:
        :rtype: NodeExecutionClosure
        """
        return cls(
            phase=p.phase,
            output_uri=p.output_uri if p.HasField("output_uri") else None,
            error=_core_execution.ExecutionError.from_flyte_idl(p.error) if p.HasField("error") else None,
            started_at=p.started_at.ToDatetime().replace(tzinfo=_pytz.UTC),
            duration=p.duration.ToTimedelta(),
        )


class NodeExecution(_common_models.FlyteIdlEntity):
    def __init__(self, id, input_uri, closure):
        """
        :param flytekit.models.core.identifier.NodeExecutionIdentifier id:
        :param Text input_uri:
        :param NodeExecutionClosure closure:
        """
        self._id = id
        self._input_uri = input_uri
        self._closure = closure

    @property
    def id(self):
        """
        :rtype: flytekit.models.core.identifier.NodeExecutionIdentifier
        """
        return self._id

    @property
    def input_uri(self):
        """
        :rtype: Text
        """
        return self._input_uri

    @property
    def closure(self):
        """
        :rtype: NodeExecutionClosure
        """
        return self._closure

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.node_execution_pb2.NodeExecution
        """
        return _node_execution_pb2.NodeExecution(
            id=self.id.to_flyte_idl(),
            input_uri=self.input_uri,
            closure=self.closure.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.admin.node_execution_pb2.NodeExecution p:
        :rtype: NodeExecution
        """
        return cls(
            id=_identifier.NodeExecutionIdentifier.from_flyte_idl(p.id),
            input_uri=p.input_uri,
            closure=NodeExecutionClosure.from_flyte_idl(p.closure),
        )



"""
Dynamic Workflows
-----------------------------
Dynamic workflows are one of the powerful aspects of Flyte. Please take a look at the :py:func:`flytekit.dynamic` documentation first to get started.


Caveats when using a dynamic workflow
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
Because of the dynamic nature of the workflow generated, it can easily be abused. Keep in mind that the workflow
that's compiled out of the decorated function needs to be processed like any other workflow. It's rare to see a
manually written workflow that has 5000 nodes for instance, but you can easily get there with a loop. Please keep,
dynamic workflows to under fifty tasks. For large-scale identical runs, we recommend the upcoming map task.

"""
import functools

from flytekit.core import task
from flytekit.core.python_function_task import PythonFunctionTask

dynamic = functools.partial(task.task, execution_mode=PythonFunctionTask.ExecutionBehavior.DYNAMIC)
dynamic.__doc__ = """
Please first see the comments for :py:func:`flytekit.task` and :py:func:`flytekit.workflow`. This ``dynamic``
concept is an amalgamation of both and enables the user to pursue some :std:ref:`pretty incredible <cookbook:sphx_glr_auto_core_control_flow_run_merge_sort.py>`
constructs.

In short, a task's function is run at execution time only, and a workflow function is run at compilation time only (local
execution notwithstanding). A dynamic workflow is modeled on the backend as a task, but at execution time, the function
body is run to produce a workflow. It is almost as if the decorator changed from ``@task`` to ``@workflow`` except workflows
cannot make use of their inputs like native Python values whereas dynamic workflows can.
The resulting workflow is passed back to the Flyte engine and is
run as a :std:ref:`subworkflow <cookbook:sphx_glr_auto_core_control_flow_subworkflows.py>`.  Simple usage

.. code-block::

    @dynamic
    def my_dynamic_subwf(a: int) -> (typing.List[str], int):
        s = []
        for i in range(a):
            s.append(t1(a=i))
        return s, 5
        
Note in the code block that we call the Python ``range`` operator on the input. This is typically not allowed in a
workflow but it is here. You can even express dependencies between tasks.

.. code-block::

    @dynamic
    def my_dynamic_subwf(a: int, b: int) -> int:
        x = t1(a=a)
        return t2(b=b, x=x)

See the :std:ref:`cookbook <cookbook:sphx_glr_auto_core_intermediate_subworkflows.py>` for a longer discussion.
"""  # noqa: W293



from google.protobuf.json_format import MessageToDict as _MessageToDict

from flytekit.common.tasks import sdk_runnable as _sdk_runnable
from flytekit.models import task as _task_models


class SdkRunnablePytorchContainer(_sdk_runnable.SdkRunnableContainer):
    @property
    def args(self):
        """
        Override args to remove the injection of command prefixes
        :rtype: list[Text]
        """
        return self._args


class SdkPyTorchTask(_sdk_runnable.SdkRunnableTask):
    def __init__(
        self,
        task_function,
        task_type,
        discovery_version,
        retries,
        interruptible,
        deprecated,
        discoverable,
        timeout,
        workers_count,
        per_replica_storage_request,
        per_replica_cpu_request,
        per_replica_gpu_request,
        per_replica_memory_request,
        per_replica_storage_limit,
        per_replica_cpu_limit,
        per_replica_gpu_limit,
        per_replica_memory_limit,
        environment,
    ):
        pytorch_job = _task_models.PyTorchJob(workers_count=workers_count).to_flyte_idl()
        super(SdkPyTorchTask, self).__init__(
            task_function=task_function,
            task_type=task_type,
            discovery_version=discovery_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=per_replica_storage_request,
            cpu_request=per_replica_cpu_request,
            gpu_request=per_replica_gpu_request,
            memory_request=per_replica_memory_request,
            storage_limit=per_replica_storage_limit,
            cpu_limit=per_replica_cpu_limit,
            gpu_limit=per_replica_gpu_limit,
            memory_limit=per_replica_memory_limit,
            discoverable=discoverable,
            timeout=timeout,
            environment=environment,
            custom=_MessageToDict(pytorch_job),
        )

    def _get_container_definition(self, **kwargs):
        """
        :rtype: SdkRunnablePytorchContainer
        """
        return super(SdkPyTorchTask, self)._get_container_definition(cls=SdkRunnablePytorchContainer, **kwargs)



import uuid as _uuid
from typing import Any, Dict, List

import six as _six
from six.moves import queue as _queue

from flytekit.common import interface as _interface
from flytekit.common import launch_plan as _launch_plan
from flytekit.common import nodes as _nodes
from flytekit.common import promise as _promise
from flytekit.common.core import identifier as _identifier
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import helpers as _type_helpers
from flytekit.common.workflow import SdkWorkflow
from flytekit.configuration import internal as _internal_config
from flytekit.models import common as _common_models
from flytekit.models import interface as _interface_models
from flytekit.models import literals as _literal_models
from flytekit.models import schedule as _schedule_models
from flytekit.models.core import identifier as _identifier_model
from flytekit.models.core import workflow as _workflow_models


# Local-only wrapper around binding data and variables. Note that the Output object used by the end user is a yet
# another layer on top of this.
class Output(object):
    def __init__(self, name, value, sdk_type=None, help=None):
        """
        :param Text name:
        :param T value:
        :param U sdk_type: If specified, the value provided must cast to this type.  Normally should be an instance of
            flytekit.common.types.base_sdk_types.FlyteSdkType.  But could also be something like:

            list[flytekit.common.types.base_sdk_types.FlyteSdkType],
            dict[flytekit.common.types.base_sdk_types.FlyteSdkType,flytekit.common.types.base_sdk_types.FlyteSdkType],
            (flytekit.common.types.base_sdk_types.FlyteSdkType, flytekit.common.types.base_sdk_types.FlyteSdkType, ...)
        """
        if sdk_type is None:
            # This syntax didn't work for some reason: sdk_type = sdk_type or Output._infer_type(value)
            sdk_type = Output._infer_type(value)
        sdk_type = _type_helpers.python_std_to_sdk_type(sdk_type)

        self._binding_data = _interface.BindingData.from_python_std(sdk_type.to_flyte_literal_type(), value)
        self._var = _interface_models.Variable(sdk_type.to_flyte_literal_type(), help or "")
        self._name = name

    def rename_and_return_reference(self, new_name):
        self._name = new_name
        return self

    @staticmethod
    def _infer_type(value):
        # TODO: Infer types
        raise NotImplementedError(
            "Currently the SDK cannot infer a workflow output type, so please use the type kwarg "
            "when instantiating an output."
        )

    @property
    def name(self):
        """
        :rtype: Text
        """
        return self._name

    @property
    def binding_data(self):
        """
        :rtype: flytekit.models.literals.BindingData
        """
        return self._binding_data

    @property
    def var(self):
        """
        :rtype: flytekit.models.interface.Variable
        """
        return self._var


class SdkRunnableWorkflow(SdkWorkflow):
    """
    Wrapper class for workflows defined using Python, written in a Flyte workflow repo. This class is misnamed. It is
    more appropriately called PythonWorkflow. The reason we are calling it SdkRunnableWorkflow instead is merely in
    keeping with the established convention in other parts of this codebase. We will likely change this naming scheme
    entirely before a 1.0 release.

    Being "runnable" or not "runnable" is not a distinction we care to make at this point. Do not read into it,
    pretend it's not there. The purpose of this class is merely to differentiate between
      i) A workflow object, as created by a user's workflow code, using the @workflow_class decorator for instance. If
         you have one of these classes, it means you have the actual Python code available in the Python process you
         are running.
      ii) The SdkWorkflow object, which represents a workflow as retrieved from Flyte Admin. Anyone with access
          to Admin, can instantiate an SdkWorkflow object by 'fetch'ing it. You don't need to have any of
          the actual code checked out. SdkWorkflow's are effectively then the control plane model of a workflow,
          represented as a Python object.
    """

    def __init__(
        self,
        inputs: List[_promise.Input],
        nodes: List[_nodes.SdkNode],
        interface,
        output_bindings,
        id=None,
        metadata=None,
        metadata_defaults=None,
        disable_default_launch_plan=False,
    ):
        """
        :param list[flytekit.common.nodes.SdkNode] nodes:
        :param flytekit.models.interface.TypedInterface interface: Defines a strongly typed interface for the
            Workflow (inputs, outputs).  This can include some optional parameters.
        :param list[flytekit.models.literals.Binding] output_bindings: A list of output bindings that specify how to construct
            workflow outputs. Bindings can pull node outputs or specify literals. All workflow outputs specified in
            the interface field must be bound
            in order for the workflow to be validated. A workflow has an implicit dependency on all of its nodes
            to execute successfully in order to bind final outputs.
        :param flytekit.models.core.identifier.Identifier id: This is an autogenerated id by the system. The id is
            globally unique across Flyte.
        :param WorkflowMetadata metadata: This contains information on how to run the workflow.
        :param flytekit.models.core.workflow.WorkflowMetadataDefaults metadata_defaults: Defaults to be passed
            to nodes contained within workflow.
        :param bool disable_default_launch_plan: Determines whether to create a default launch plan for the workflow.
        """
        # Save the promise.Input objects for future use.
        self._user_inputs = inputs

        # Set optional settings
        id = (
            id
            if id is not None
            else _identifier.Identifier(
                _identifier_model.ResourceType.WORKFLOW,
                _internal_config.PROJECT.get(),
                _internal_config.DOMAIN.get(),
                _uuid.uuid4().hex,
                _internal_config.VERSION.get(),
            )
        )
        metadata = metadata if metadata is not None else _workflow_models.WorkflowMetadata()
        metadata_defaults = (
            metadata_defaults if metadata_defaults is not None else _workflow_models.WorkflowMetadataDefaults()
        )

        super(SdkRunnableWorkflow, self).__init__(
            nodes=nodes,
            interface=interface,
            output_bindings=output_bindings,
            id=id,
            metadata=metadata,
            metadata_defaults=metadata_defaults,
        )

        # Set this last as it's set in constructor
        self._upstream_entities = set(n.executable_sdk_object for n in nodes)
        self._should_create_default_launch_plan = not disable_default_launch_plan

    @property
    def should_create_default_launch_plan(self):
        """
        Determines whether registration flow should create a default launch plan for this workflow or not.
        :rtype: bool
        """
        return self._should_create_default_launch_plan

    def __call__(self, *args, **input_map):
        # Take the default values from the Inputs
        compiled_inputs = {v.name: v.sdk_default for v in self.user_inputs if not v.sdk_required}
        compiled_inputs.update(input_map)

        return super().__call__(*args, **compiled_inputs)

    @classmethod
    def construct_from_class_definition(
        cls,
        inputs: List[_promise.Input],
        outputs: List[Output],
        nodes: List[_nodes.SdkNode],
        metadata: _workflow_models.WorkflowMetadata = None,
        metadata_defaults: _workflow_models.WorkflowMetadataDefaults = None,
        disable_default_launch_plan: bool = False,
    ) -> "SdkRunnableWorkflow":
        """
        This constructor is here to provide backwards-compatibility for class-defined Workflows

        :param list[flytekit.common.promise.Input] inputs:
        :param list[Output] outputs:
        :param list[flytekit.common.nodes.SdkNode] nodes:
        :param WorkflowMetadata metadata: This contains information on how to run the workflow.
        :param flytekit.models.core.workflow.WorkflowMetadataDefaults metadata_defaults: Defaults to be passed
            to nodes contained within workflow.
        :param bool disable_default_launch_plan: Determines whether to create a default launch plan for the workflow or not.

        :rtype: SdkRunnableWorkflow
        """
        for n in nodes:
            for upstream in n.upstream_nodes:
                if upstream.id is None:
                    raise _user_exceptions.FlyteAssertion(
                        "Some nodes contained in the workflow were not found in the workflow description.  Please "
                        "ensure all nodes are either assigned to attributes within the class or an element in a "
                        "list, dict, or tuple which is stored as an attribute in the class."
                    )

        id = _identifier.Identifier(
            _identifier_model.ResourceType.WORKFLOW,
            _internal_config.PROJECT.get(),
            _internal_config.DOMAIN.get(),
            _uuid.uuid4().hex,
            _internal_config.VERSION.get(),
        )
        interface = _interface.TypedInterface({v.name: v.var for v in inputs}, {v.name: v.var for v in outputs})

        output_bindings = [_literal_models.Binding(v.name, v.binding_data) for v in outputs]

        return cls(
            inputs=inputs,
            nodes=nodes,
            interface=interface,
            output_bindings=output_bindings,
            id=id,
            metadata=metadata,
            metadata_defaults=metadata_defaults,
            disable_default_launch_plan=disable_default_launch_plan,
        )

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, new_id):
        self._id = new_id

    @property
    def user_inputs(self) -> List[_promise.Input]:
        """
        :rtype: list[flytekit.common.promise.Input]
        """
        return self._user_inputs

    def create_launch_plan(
        self,
        default_inputs: Dict[str, _promise.Input] = None,
        fixed_inputs: Dict[str, Any] = None,
        schedule: _schedule_models.Schedule = None,
        role: str = None,
        notifications: List[_common_models.Notification] = None,
        labels: _common_models.Labels = None,
        annotations: _common_models.Annotations = None,
        assumable_iam_role: str = None,
        kubernetes_service_account: str = None,
        raw_output_data_prefix: str = None,
    ):
        """
        This method will create a launch plan object that can execute this workflow.
        :param dict[Text,flytekit.common.promise.Input] default_inputs:
        :param dict[Text,T] fixed_inputs:
        :param flytekit.models.schedule.Schedule schedule: A schedule on which to execute this launch plan.
        :param Text role: Deprecated. Use assumable_iam_role instead.
        :param list[flytekit.models.common.Notification] notifications: A list of notifications to enact by default for
        this launch plan.
        :param flytekit.models.common.Labels labels:
        :param flytekit.models.common.Annotations annotations:
        :param cls: This parameter can be used by users to define an extension of a launch plan to instantiate.  The
        class provided should be a subclass of flytekit.common.launch_plan.SdkLaunchPlan.
        :param Text assumable_iam_role: The IAM role to execute the workflow with.
        :param Text kubernetes_service_account: The kubernetes service account to execute the workflow with.
        :param Text raw_output_data_prefix: Bucket for offloaded data
        :rtype: flytekit.common.launch_plan.SdkRunnableLaunchPlan
        """
        # TODO: Actually ensure the parameters conform.
        if role and (assumable_iam_role or kubernetes_service_account):
            raise ValueError("Cannot set both role and auth. Role is deprecated, use auth instead.")
        fixed_inputs = fixed_inputs or {}
        merged_default_inputs = {v.name: v for v in self.user_inputs if v.name not in fixed_inputs}
        merged_default_inputs.update(default_inputs or {})

        if role:
            assumable_iam_role = role  # For backwards compatibility
        auth_role = _common_models.AuthRole(
            assumable_iam_role=assumable_iam_role,
            kubernetes_service_account=kubernetes_service_account,
        )

        raw_output_config = _common_models.RawOutputDataConfig(raw_output_data_prefix or "")

        return _launch_plan.SdkRunnableLaunchPlan(
            sdk_workflow=self,
            default_inputs={
                k: user_input.rename_and_return_reference(k) for k, user_input in _six.iteritems(merged_default_inputs)
            },
            fixed_inputs=fixed_inputs,
            schedule=schedule,
            notifications=notifications,
            labels=labels,
            annotations=annotations,
            auth_role=auth_role,
            raw_output_data_config=raw_output_config,
        )


def build_sdk_workflow_from_metaclass(metaclass, on_failure=None, disable_default_launch_plan=False, cls=None):
    """
    :param T metaclass: This is the user-defined workflow class, prior to decoration.
    :param on_failure flytekit.models.core.workflow.WorkflowMetadata.OnFailurePolicy: [Optional] The execution policy
    when the workflow detects a failure.
    :param bool disable_default_launch_plan: Determines whether to create a default launch plan for the workflow or not.
    :param cls: This is the class that will be instantiated from the inputs, outputs, and nodes.  This will be used
        by users extending the base Flyte programming model. If set, it must be a subclass of PythonWorkflow.

    :rtype: SdkRunnableWorkflow
    """
    inputs, outputs, nodes = _discover_workflow_components(metaclass)
    metadata = _workflow_models.WorkflowMetadata(on_failure=on_failure if on_failure else None)

    return (cls or SdkRunnableWorkflow).construct_from_class_definition(
        inputs=[i for i in sorted(inputs, key=lambda x: x.name)],
        outputs=[o for o in sorted(outputs, key=lambda x: x.name)],
        nodes=[n for n in sorted(nodes, key=lambda x: x.id)],
        metadata=metadata,
        disable_default_launch_plan=disable_default_launch_plan,
    )


def _discover_workflow_components(workflow_class):
    """
    This task iterates over the attributes of a user-defined class in order to return a list of inputs, outputs and
    nodes.
    :param class workflow_class: User-defined class with task instances as attributes.
    :rtype: (list[flytekit.common.promise.Input], list[Output], list[flytekit.common.nodes.SdkNode])
    """

    inputs = []
    outputs = []
    nodes = []

    to_visit_objs = _queue.Queue()
    top_level_attributes = set()
    for attribute_name in dir(workflow_class):
        to_visit_objs.put((attribute_name, getattr(workflow_class, attribute_name)))
        top_level_attributes.add(attribute_name)

    # For all task instances defined within the workflow, bind them to this specific workflow and hook-up to the
    # engine (when available)
    visited_obj_ids = set()
    while not to_visit_objs.empty():
        attribute_name, current_obj = to_visit_objs.get()

        current_obj_id = id(current_obj)
        if current_obj_id in visited_obj_ids:
            continue
        visited_obj_ids.add(current_obj_id)

        if isinstance(current_obj, _nodes.SdkNode):
            # TODO: If an attribute name is on the form node_name[index], the resulting
            # node name might not be correct.
            nodes.append(current_obj.assign_id_and_return(attribute_name))
        elif isinstance(current_obj, _promise.Input):
            if attribute_name is None or attribute_name not in top_level_attributes:
                raise _user_exceptions.FlyteValueException(
                    attribute_name,
                    "Detected workflow input specified outside of top level.",
                )
            inputs.append(current_obj.rename_and_return_reference(attribute_name))
        elif isinstance(current_obj, Output):
            if attribute_name is None or attribute_name not in top_level_attributes:
                raise _user_exceptions.FlyteValueException(
                    attribute_name,
                    "Detected workflow output specified outside of top level.",
                )
            outputs.append(current_obj.rename_and_return_reference(attribute_name))
        elif isinstance(current_obj, list) or isinstance(current_obj, set) or isinstance(current_obj, tuple):
            for idx, value in enumerate(current_obj):
                to_visit_objs.put((_assign_indexed_attribute_name(attribute_name, idx), value))
        elif isinstance(current_obj, dict):
            # Visit dictionary keys.
            for key in current_obj.keys():
                to_visit_objs.put((_assign_indexed_attribute_name(attribute_name, key), key))
            # Visit dictionary values.
            for key, value in _six.iteritems(current_obj):
                to_visit_objs.put((_assign_indexed_attribute_name(attribute_name, key), value))
    return inputs, outputs, nodes


def _assign_indexed_attribute_name(attribute_name, index):
    return "{}[{}]".format(attribute_name, index)



from flyteidl.core import condition_pb2 as _condition

from flytekit.models import common as _common
from flytekit.models import literals as _literals


class ComparisonExpression(_common.FlyteIdlEntity):
    class Operator(object):
        """
        Binary Operator for each expression
        """

        EQ = _condition.ComparisonExpression.EQ
        NEQ = _condition.ComparisonExpression.NEQ
        GT = _condition.ComparisonExpression.GT
        GTE = _condition.ComparisonExpression.GTE
        LT = _condition.ComparisonExpression.LT
        LTE = _condition.ComparisonExpression.LTE

    def __init__(self, operator, left_value, right_value):
        """
        Defines a 2-level tree where the root is a comparison operator and Operands are primitives or known variables.
        Each expression results in a boolean result.

        :param Operator operator:
        :param Operand left_value:
        :param Operand right_value:
        """
        self._operator = operator
        self._left_value = left_value
        self._right_value = right_value

    @property
    def operator(self):
        """
        Gets the operator representing this comparison expression.
        :rtype: ComparisonExpression.Operator
        """
        return self._operator

    @property
    def left_value(self):
        """
        Gets the left value for the comparison expression.
        :rtype: Operand
        """
        return self._left_value

    @property
    def right_value(self):
        """
        Gets the right value for the comparison expression.
        :rtype: Operand
        """
        return self._right_value

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.condition_pb2.ComparisonExpression
        """
        return _condition.ComparisonExpression(
            operator=self.operator,
            left_value=self.left_value.to_flyte_idl(),
            right_value=self.right_value.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        return cls(
            operator=pb2_object.operator,
            left_value=Operand.from_flyte_idl(pb2_object.left_value),
            right_value=Operand.from_flyte_idl(pb2_object.right_value),
        )


class ConjunctionExpression(_common.FlyteIdlEntity):
    class LogicalOperator(object):
        AND = _condition.ConjunctionExpression.AND
        OR = _condition.ConjunctionExpression.OR

    def __init__(self, operator, left_expression, right_expression):
        """
        Defines a conjunction expression of two boolean expressions.
        :param LogicalOperator operator:
        :param BooleanExpression left_expression:
        :param BooleanExpression right_expression:
        """

        self._operator = operator
        self._left_expression = left_expression
        self._right_expression = right_expression

    @property
    def operator(self):
        """
        Gets the operator representing this conjunction expression.
        :rtype: ConjunctionExpression.LogicalOperator
        """
        return self._operator

    @property
    def left_expression(self):
        """
        Gets the left value for the conjunction expression.
        :rtype: Operand
        """
        return self._left_expression

    @property
    def right_expression(self):
        """
        Gets the right value for the conjunction expression.
        :rtype: Operand
        """
        return self._right_expression

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.condition_pb2.ConjunctionExpression
        """
        return _condition.ConjunctionExpression(
            operator=self.operator,
            left_expression=self.left_expression.to_flyte_idl(),
            right_expression=self.right_expression.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        return cls(
            operator=pb2_object.operator,
            left_expression=BooleanExpression.from_flyte_idl(pb2_object.left_expression),
            right_expression=BooleanExpression.from_flyte_idl(pb2_object.right_expression),
        )


class Operand(_common.FlyteIdlEntity):
    def __init__(self, primitive=None, var=None):
        """
        Defines an operand to a comparison expression.
        :param flytekit.models.literals.Primitive primitive:
        :param Text var:
        """

        self._primitive = primitive
        self._var = var

    @property
    def primitive(self):
        """
        :rtype: flytekit.models.literals.Primitive
        """

        return self._primitive

    @property
    def var(self):
        """
        :rtype: Text
        """

        return self._var

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.condition_pb2.Operand
        """
        return _condition.Operand(
            primitive=self.primitive.to_flyte_idl() if self.primitive else None,
            var=self.var if self.var else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        return cls(
            primitive=_literals.Primitive.from_flyte_idl(pb2_object.primitive)
            if pb2_object.HasField("primitive")
            else None,
            var=pb2_object.var if pb2_object.HasField("var") else None,
        )


class BooleanExpression(_common.FlyteIdlEntity):
    def __init__(self, conjunction=None, comparison=None):
        """
        Defines a boolean expression tree. It can be a simple or a conjunction expression.
        Multiple expressions can be combined using a conjunction or a disjunction to result in a final boolean result.

        :param ConjunctionExpression conjunction:
        :param ComparisonExpression comparison:
        """

        self._conjunction = conjunction
        self._comparison = comparison

    @property
    def conjunction(self):
        """
        Conjunction expression or None if not set.
        :rtype: ConjunctionExpression
        """
        return self._conjunction

    @property
    def comparison(self):
        """
        Comparison expression or None if not set.
        :rtype: ComparisonExpression
        """
        return self._comparison

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.condition_pb2.BooleanExpression
        """
        return _condition.BooleanExpression(
            conjunction=self.conjunction.to_flyte_idl() if self.conjunction else None,
            comparison=self.comparison.to_flyte_idl() if self.comparison else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        return cls(
            conjunction=ConjunctionExpression.from_flyte_idl(pb2_object.conjunction)
            if pb2_object.HasField("conjunction")
            else None,
            comparison=ComparisonExpression.from_flyte_idl(pb2_object.comparison)
            if pb2_object.HasField("comparison")
            else None,
        )



from __future__ import annotations

from typing import Any, List

from flytekit.common.utils import _dnsify
from flytekit.models import literals as _literal_models
from flytekit.models.core import workflow as _workflow_model


class Node(object):
    """
    This class will hold all the things necessary to make an SdkNode but we won't make one until we know things like
    ID, which from the registration step
    """

    def __init__(
        self,
        id: str,
        metadata: _workflow_model.NodeMetadata,
        bindings: List[_literal_models.Binding],
        upstream_nodes: List[Node],
        flyte_entity: Any,
    ):
        self._id = _dnsify(id)
        self._metadata = metadata
        self._bindings = bindings
        self._upstream_nodes = upstream_nodes
        self._flyte_entity = flyte_entity
        self._aliases: _workflow_model.Alias = None
        self._outputs = None

    def runs_before(self, other: Node):
        """
        This is typically something we shouldn't do. This modifies an attribute of the other instance rather than
        self. But it's done so only because we wanted this English function to be the same as the shift function.
        That is, calling node_1.runs_before(node_2) and node_1 >> node_2 are the same. The shift operator going the
        other direction is not implemented to further avoid confusion. Right shift was picked rather than left shift
        because that's what most users are familiar with.
        """
        if self not in other._upstream_nodes:
            other._upstream_nodes.append(self)

    def __rshift__(self, other: Node):
        self.runs_before(other)

    @property
    def outputs(self):
        if self._outputs is None:
            raise AssertionError("Cannot use outputs with all Nodes, node must've been created from create_node()")
        return self._outputs

    @property
    def id(self) -> str:
        return self._id

    @property
    def bindings(self) -> List[_literal_models.Binding]:
        return self._bindings

    @property
    def upstream_nodes(self) -> List[Node]:
        return self._upstream_nodes

    @property
    def flyte_entity(self) -> Any:
        return self._flyte_entity

    def with_overrides(self, *args, **kwargs):
        if "node_name" in kwargs:
            self._id = kwargs["node_name"]
        if "aliases" in kwargs:
            alias_dict = kwargs["aliases"]
            if not isinstance(alias_dict, dict):
                raise AssertionError("Aliases should be specified as dict[str, str]")
            self._aliases = []
            for k, v in alias_dict.items():
                self._aliases.append(_workflow_model.Alias(var=k, alias=v))
        return self



from flytekit.configuration import common as _common_config

HOST = _common_config.FlyteStringConfigurationEntry("statsd", "host", default="localhost")
PORT = _common_config.FlyteIntegerConfigurationEntry("statsd", "port", default=8125)
DISABLED = _common_config.FlyteBoolConfigurationEntry("statsd", "disabled", default=False)



import base64 as _base64
from typing import Type, Union

import six as _six
from google.protobuf import reflection as _proto_reflection
from google.protobuf.json_format import Error
from google.protobuf.json_format import MessageToDict as _MessageToDict
from google.protobuf.json_format import ParseDict as _ParseDict
from google.protobuf.reflection import GeneratedProtocolMessageType
from google.protobuf.struct_pb2 import Struct

from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import base_sdk_types as _base_sdk_types
from flytekit.models import literals as _literals
from flytekit.models import types as _idl_types
from flytekit.models.common import FlyteIdlEntity, FlyteType
from flytekit.models.types import LiteralType

ProtobufT = Type[_proto_reflection.GeneratedProtocolMessageType]


class ProtobufType(_base_sdk_types.FlyteSdkType):
    _pb_type = Struct

    @property
    def pb_type(cls) -> GeneratedProtocolMessageType:
        """
        :rtype: GeneratedProtocolMessageType
        """
        return cls._pb_type

    @property
    def descriptor(cls):
        """
        :rtype: Text
        """
        return "{}.{}".format(cls.pb_type.__module__, cls.pb_type.__name__)

    @property
    def tag(cls):
        """
        :rtype: Text
        """
        return "{}{}".format(Protobuf.TAG_PREFIX, cls.descriptor)


class Protobuf(_base_sdk_types.FlyteSdkValue, metaclass=ProtobufType):
    PB_FIELD_KEY = "pb_type"
    TAG_PREFIX = "{}=".format(PB_FIELD_KEY)

    def __init__(self, pb_object: Union[GeneratedProtocolMessageType, FlyteIdlEntity]):
        """
        :param Union[T, FlyteIdlEntity] pb_object:
        """
        v = pb_object
        # This section converts an existing proto object (or a subclass of) to the right type expected by this instance
        # of GenericProto. GenericProto can be used with any protobuf type (not restricted to FlyteType). This makes it
        # a bit tricky to figure out the right version of the underlying raw proto class to use to populate the final
        # struct.
        # If the provided object has to_flyte_idl(), call it to produce a raw proto.
        if isinstance(pb_object, FlyteIdlEntity):
            v = pb_object.to_flyte_idl()

        # A check to ensure the raw proto (v) is of the correct expected type. This also performs one final attempt to
        # convert it to the correct type by leveraging from_flyte_idl (implemented by all FlyteTypes) in case this class
        # is initialized with one.
        expected_type = type(self).pb_type
        if expected_type != type(v) and expected_type != type(pb_object):
            if isinstance(type(self).pb_type, FlyteType):
                v = expected_type.from_flyte_idl(v).to_flyte_idl()
            else:
                raise _user_exceptions.FlyteTypeException(
                    received_type=type(pb_object), expected_type=expected_type, received_value=pb_object
                )
        data = v.SerializeToString()
        super(Protobuf, self).__init__(
            scalar=_literals.Scalar(
                binary=_literals.Binary(value=bytes(data) if _six.PY2 else data, tag=type(self).tag)
            )
        )

    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value: b64 encoded string of bytes
        :rtype: Protobuf
        """
        try:
            decoded = _base64.b64decode(string_value)
        except TypeError:
            raise _user_exceptions.FlyteValueException(string_value, "The string is not valid base64-encoded.")
        pb_obj = cls.pb_type()
        pb_obj.ParseFromString(decoded)
        return cls(pb_obj)

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return isinstance(other, ProtobufType) and other.pb_type is cls.pb_type

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: _base_sdk_types.FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif isinstance(t_value, cls.pb_type) or isinstance(t_value, FlyteIdlEntity):
            return cls(t_value)
        else:
            raise _user_exceptions.FlyteTypeException(type(t_value), cls.pb_type, received_value=t_value)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.BINARY, metadata={cls.PB_FIELD_KEY: cls.descriptor})

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Protobuf
        """
        if literal_model.scalar.binary.tag != cls.tag:
            raise _user_exceptions.FlyteTypeException(
                literal_model.scalar.binary.tag,
                cls.pb_type,
                received_value=_base64.b64encode(literal_model.scalar.binary.value),
                additional_msg="Can not deserialize as proto tags don't match.",
            )
        pb_obj = cls.pb_type()
        pb_obj.ParseFromString(literal_model.scalar.binary.value)
        return cls(pb_obj)

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Types.Proto({})".format(cls.descriptor)

    def to_python_std(self):
        """
        :returns: The protobuf object as defined by the user.
        :rtype: T
        """
        pb_obj = type(self).pb_type()
        pb_obj.ParseFromString(self.scalar.binary.value)
        return pb_obj

    def short_string(self):
        """
        :rtype: Text
        """
        return "{}".format(self.to_python_std())


def create_protobuf(pb_type: Type[GeneratedProtocolMessageType]) -> Type[Protobuf]:
    """
    :param Type[GeneratedProtocolMessageType] pb_type:
    :rtype: Type[Protobuf]
    """
    if not isinstance(pb_type, _proto_reflection.GeneratedProtocolMessageType):
        raise _user_exceptions.FlyteTypeException(
            expected_type=_proto_reflection.GeneratedProtocolMessageType,
            received_type=type(pb_type),
            received_value=pb_type,
        )

    class _Protobuf(Protobuf):
        _pb_type = pb_type

    return _Protobuf


class GenericProtobuf(_base_sdk_types.FlyteSdkValue, metaclass=ProtobufType):
    PB_FIELD_KEY = "pb_type"
    TAG_PREFIX = "{}=".format(PB_FIELD_KEY)

    def __init__(self, pb_object: Union[GeneratedProtocolMessageType, FlyteIdlEntity]):
        """
        :param Union[T, FlyteIdlEntity] pb_object:
        """
        struct = Struct()
        v = pb_object

        # This section converts an existing proto object (or a subclass of) to the right type expected by this instance
        # of GenericProto. GenericProto can be used with any protobuf type (not restricted to FlyteType). This makes it
        # a bit tricky to figure out the right version of the underlying raw proto class to use to populate the final
        # struct.
        # If the provided object has to_flyte_idl(), call it to produce a raw proto.
        if isinstance(pb_object, FlyteIdlEntity):
            v = pb_object.to_flyte_idl()

        # A check to ensure the raw proto (v) is of the correct expected type. This also performs one final attempt to
        # convert it to the correct type by leveraging from_flyte_idl (implemented by all FlyteTypes) in case this class
        # is initialized with one.
        expected_type = type(self).pb_type
        if expected_type != type(v) and expected_type != type(pb_object):
            if isinstance(type(self).pb_type, FlyteType):
                v = expected_type.from_flyte_idl(v).to_flyte_idl()
            else:
                raise _user_exceptions.FlyteTypeException(
                    received_type=type(pb_object), expected_type=expected_type, received_value=pb_object
                )

        struct.update(_MessageToDict(v))
        super().__init__(scalar=_literals.Scalar(generic=struct))

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return isinstance(other, ProtobufType) and other.pb_type is cls.pb_type

    @classmethod
    def from_python_std(cls, t_value: Union[GeneratedProtocolMessageType, FlyteIdlEntity]):
        """
        :param Union[T, FlyteIdlEntity] t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: _base_sdk_types.FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif isinstance(t_value, cls.pb_type) or isinstance(t_value, FlyteIdlEntity):
            return cls(t_value)
        else:
            raise _user_exceptions.FlyteTypeException(type(t_value), cls.pb_type, received_value=t_value)

    @classmethod
    def to_flyte_literal_type(cls) -> LiteralType:
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.STRUCT, metadata={cls.PB_FIELD_KEY: cls.descriptor})

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Protobuf
        """
        pb_obj = cls.pb_type()
        try:
            dictionary = _MessageToDict(literal_model.scalar.generic)
            pb_obj = _ParseDict(dictionary, pb_obj)
        except Error as err:
            raise _user_exceptions.FlyteTypeException(
                received_type="generic",
                expected_type=cls.pb_type,
                received_value=_base64.b64encode(literal_model.scalar.generic),
                additional_msg=f"Can not deserialize. Error: {err.__str__()}",
            )

        return cls(pb_obj)

    @classmethod
    def short_class_string(cls) -> str:
        """
        :rtype: Text
        """
        return "Types.GenericProto({})".format(cls.descriptor)

    def to_python_std(self):
        """
        :returns: The protobuf object as defined by the user.
        :rtype: T
        """
        pb_obj = type(self).pb_type()
        try:
            dictionary = _MessageToDict(self.scalar.generic)
            pb_obj = _ParseDict(dictionary, pb_obj)
        except Error as err:
            raise _user_exceptions.FlyteTypeException(
                received_type="generic",
                expected_type=type(self).pb_type,
                received_value=_base64.b64encode(self.scalar.generic),
                additional_msg=f"Can not deserialize. Error: {err.__str__()}",
            )
        return pb_obj

    def short_string(self) -> str:
        """
        :rtype: Text
        """
        return "{}".format(self.to_python_std())


def create_generic(pb_type: Type[GeneratedProtocolMessageType]) -> Type[GenericProtobuf]:
    """
    Creates a generic protobuf type that represents protobuf type ProtobufT and that will get serialized into a struct.

    :param Type[GeneratedProtocolMessageType] pb_type:
    :rtype: Type[GenericProtobuf]
    """
    if not isinstance(pb_type, _proto_reflection.GeneratedProtocolMessageType) and not issubclass(
        pb_type, FlyteIdlEntity
    ):
        raise _user_exceptions.FlyteTypeException(
            expected_type=_proto_reflection.GeneratedProtocolMessageType,
            received_type=type(pb_type),
            received_value=pb_type,
        )

    class _Protobuf(GenericProtobuf):
        _pb_type = pb_type

    return _Protobuf



from flyteidl.core import dynamic_job_pb2 as _dynamic_job

from flytekit.models import common as _common
from flytekit.models import literals as _literals
from flytekit.models import task as _task
from flytekit.models.core import workflow as _workflow


class DynamicJobSpec(_common.FlyteIdlEntity):
    def __init__(self, tasks, nodes, min_successes, outputs, subworkflows):
        """
        Initializes a new FutureTaskDocument.

        :param list[flytekit.models.task.TaskTemplate] tasks: A collection of unique tasks to execute.
        :param list[flytekit.models.core.workflow.Node] nodes: A collection of task nodes.
        :param int min_successes: An absolute number of the minimum number of successful completions of subtasks. As
            soon as this criteria is met, the future job will be marked as successful and outputs will be computed.
        :param list[flytekit.models.literals.Binding] outputs: Describes how to bind the final output of the future
            task from the
            outputs of executed nodes. The referenced ids in bindings should have the generated id for the subtask.
        :param list[flytekit.models.workflow.WorkflowTemplate] subworkflows: A complete list of task specs referenced
            in nodes.
        """

        self._tasks = tasks
        self._nodes = nodes
        self._min_successes = min_successes
        self._outputs = outputs
        self._subworkflows = subworkflows

    @property
    def tasks(self):
        """
        A collection of tasks to execute.
        :rtype: list[_task.TaskTemplate]
        """
        return self._tasks

    @property
    def nodes(self):
        """
        A collection of dynamic nodes.
        :rtype: list[_workflow.Node]
        """
        return self._nodes

    @property
    def min_successes(self):
        """
        An absolute number of the minimum number of successful completions of subtasks. As
            soon as this criteria is met, the future job will be marked as successful and outputs will be computed.
        :rtype: int
        """
        return self._min_successes

    @property
    def outputs(self):
        """
        Describes how to bind the final output of the future task from the outputs of executed nodes.
            The referenced ids in bindings should have the generated id for the subtask.
        :rtype: list[flytekit.models.literals.Binding]
        """
        return self._outputs

    @property
    def subworkflows(self):
        """
        A collection of subworkflows to execute.
        :rtype: list[flytekit.models.core.workflow.WorkflowTemplate]
        """
        return self._subworkflows

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.dynamic_job.DynamicJobSpec
        """
        return _dynamic_job.DynamicJobSpec(
            tasks=[task.to_flyte_idl() for task in self.tasks] if self.tasks else None,
            nodes=[node.to_flyte_idl() for node in self.nodes] if self.nodes else None,
            min_successes=self.min_successes,
            outputs=[output.to_flyte_idl() for output in self.outputs],
            subworkflows=[workflow.to_flyte_idl() for workflow in self.subworkflows],
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.dynamic_job_pb2.DynamicJobSpec pb2_object:
        :return: DynamicJobSpec
        """
        return cls(
            tasks=[_task.TaskTemplate.from_flyte_idl(task) for task in pb2_object.tasks] if pb2_object.tasks else None,
            nodes=[_workflow.Node.from_flyte_idl(n) for n in pb2_object.nodes],
            min_successes=pb2_object.min_successes,
            outputs=[_literals.Binding.from_flyte_idl(output) for output in pb2_object.outputs]
            if pb2_object.outputs
            else None,
            subworkflows=[_workflow.WorkflowTemplate.from_flyte_idl(w) for w in pb2_object.subworkflows],
        )



from dataclasses import dataclass


@dataclass
class Resources(object):
    """
    This class is used to specify both resource requests and resource limits. ::

        Resources(cpu="1", mem="2048")  # This is 1 CPU and 2 KB of memory
        Resources(cpu="100m", mem="2Gi")  # This is 1/10th of a CPU and 2 gigabytes of memory

    .. note::
        Storage is not currently supported on the Flyte backend.

    Please see the :std:ref:`User Guide <cookbook:sphx_glr_auto_deployment_workflow_customizing_resources.py>` for detailed examples.
    Also refer to the `K8s conventions. <https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/#resource-units-in-kubernetes>`__
    """

    cpu: str = None
    mem: str = None
    gpu: str = None
    storage: str = None


@dataclass
class ResourceSpec(object):
    requests: Resources = None
    limits: Resources = None



import datetime as _datetime

from google.protobuf.json_format import MessageToDict

from flytekit import __version__
from flytekit.common import interface as _interface
from flytekit.common.constants import SdkTaskType
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.tasks import task as _sdk_task
from flytekit.models import interface as _interface_model
from flytekit.models import literals as _literal_models
from flytekit.models import task as _task_models
from flytekit.models import types as _idl_types
from flytekit.models.core import types as _core_types
from flytekit.models.sagemaker import training_job as _training_job_models


def _content_type_to_blob_format(content_type: _training_job_models) -> str:
    if content_type == _training_job_models.InputContentType.TEXT_CSV:
        return "csv"
    else:
        raise _user_exceptions.FlyteValueException("Unsupported InputContentType: {}".format(content_type))


class SdkBuiltinAlgorithmTrainingJobTask(_sdk_task.SdkTask):
    def __init__(
        self,
        training_job_resource_config: _training_job_models.TrainingJobResourceConfig,
        algorithm_specification: _training_job_models.AlgorithmSpecification,
        retries: int = 0,
        cacheable: bool = False,
        cache_version: str = "",
    ):
        """

        :param training_job_resource_config: The options to configure the training job
        :param algorithm_specification: The options to configure the target algorithm of the training
        :param retries: Number of retries to attempt
        :param cacheable: The flag to set if the user wants the output of the task execution to be cached
        :param cache_version: String describing the caching version for task discovery purposes
        """
        # Use the training job model as a measure of type checking
        self._training_job_model = _training_job_models.TrainingJob(
            algorithm_specification=algorithm_specification,
            training_job_resource_config=training_job_resource_config,
        )

        # Setting flyte-level timeout to 0, and let SageMaker takes the StoppingCondition and terminate the training
        # job gracefully
        timeout = _datetime.timedelta(seconds=0)

        super(SdkBuiltinAlgorithmTrainingJobTask, self).__init__(
            type=SdkTaskType.SAGEMAKER_TRAINING_JOB_TASK,
            metadata=_task_models.TaskMetadata(
                runtime=_task_models.RuntimeMetadata(
                    type=_task_models.RuntimeMetadata.RuntimeType.FLYTE_SDK,
                    version=__version__,
                    flavor="sagemaker",
                ),
                discoverable=cacheable,
                timeout=timeout,
                retries=_literal_models.RetryStrategy(retries=retries),
                interruptible=False,
                discovery_version=cache_version,
                deprecated_error_message="",
            ),
            interface=_interface.TypedInterface(
                inputs={
                    "static_hyperparameters": _interface_model.Variable(
                        type=_idl_types.LiteralType(simple=_idl_types.SimpleType.STRUCT),
                        description="",
                    ),
                    "train": _interface_model.Variable(
                        type=_idl_types.LiteralType(
                            blob=_core_types.BlobType(
                                format=_content_type_to_blob_format(algorithm_specification.input_content_type),
                                dimensionality=_core_types.BlobType.BlobDimensionality.MULTIPART,
                            ),
                        ),
                        description="",
                    ),
                    "validation": _interface_model.Variable(
                        type=_idl_types.LiteralType(
                            blob=_core_types.BlobType(
                                format=_content_type_to_blob_format(algorithm_specification.input_content_type),
                                dimensionality=_core_types.BlobType.BlobDimensionality.MULTIPART,
                            ),
                        ),
                        description="",
                    ),
                },
                outputs={
                    "model": _interface_model.Variable(
                        type=_idl_types.LiteralType(
                            blob=_core_types.BlobType(
                                format="",
                                dimensionality=_core_types.BlobType.BlobDimensionality.SINGLE,
                            )
                        ),
                        description="",
                    )
                },
            ),
            custom=MessageToDict(self._training_job_model.to_flyte_idl()),
        )

    @property
    def training_job_model(self) -> _training_job_models.TrainingJob:
        return self._training_job_model



"""
Flytekit Schema Type (:mod:`flytekit.types.schema`)
==========================================================
.. currentmodule:: flytekit.types.schema

.. autosummary::
   :toctree: generated/

   SchemaFormat
   FlyteSchema
   FlyteSchema.open
"""

from .types import (
    FlyteSchema,
    LocalIOSchemaReader,
    LocalIOSchemaWriter,
    SchemaEngine,
    SchemaFormat,
    SchemaHandler,
    SchemaOpenMode,
    SchemaReader,
    SchemaWriter,
)
from .types_pandas import PandasSchemaReader, PandasSchemaWriter



from google.protobuf.json_format import MessageToDict as _MessageToDict

from flytekit.common.tasks import sdk_runnable as _sdk_runnable
from flytekit.models import task as _task_models


class SdkRunnableTensorflowContainer(_sdk_runnable.SdkRunnableContainer):
    @property
    def args(self):
        """
        Override args to remove the injection of command prefixes
        :rtype: list[Text]
        """
        return self._args


class SdkTensorFlowTask(_sdk_runnable.SdkRunnableTask):
    def __init__(
        self,
        task_function,
        task_type,
        cache_version,
        retries,
        interruptible,
        deprecated,
        cache,
        timeout,
        workers_count,
        ps_replicas_count,
        chief_replicas_count,
        per_replica_storage_request,
        per_replica_cpu_request,
        per_replica_gpu_request,
        per_replica_memory_request,
        per_replica_storage_limit,
        per_replica_cpu_limit,
        per_replica_gpu_limit,
        per_replica_memory_limit,
        environment,
    ):
        tensorflow_job = _task_models.TensorFlowJob(
            workers_count=workers_count, ps_replicas_count=ps_replicas_count, chief_replicas_count=chief_replicas_count
        ).to_flyte_idl()
        super(SdkTensorFlowTask, self).__init__(
            task_function=task_function,
            task_type=task_type,
            discovery_version=cache_version,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=per_replica_storage_request,
            cpu_request=per_replica_cpu_request,
            gpu_request=per_replica_gpu_request,
            memory_request=per_replica_memory_request,
            storage_limit=per_replica_storage_limit,
            cpu_limit=per_replica_cpu_limit,
            gpu_limit=per_replica_gpu_limit,
            memory_limit=per_replica_memory_limit,
            discoverable=cache,
            timeout=timeout,
            environment=environment,
            custom=_MessageToDict(tensorflow_job),
        )

    def _get_container_definition(self, **kwargs):
        """
        :rtype: SdkRunnableTensorflowContainer
        """
        return super(SdkTensorFlowTask, self)._get_container_definition(cls=SdkRunnableTensorflowContainer, **kwargs)



## Todo - change this to qubole_presto once Luis's PR get's merged
# from flyteidl.plugins import qubole_presto as _qubole
from flyteidl.plugins import presto_pb2 as _presto

from flytekit.models import common as _common


class PrestoQuery(_common.FlyteIdlEntity):
    def __init__(self, routing_group, catalog, schema, statement):
        """
        Initializes a new PrestoQuery.

        :param string routing_group:
        :param string catalog:
        :param string schema:
        :param string statement:

        """
        self._routing_group = routing_group
        self._catalog = catalog
        self._schema = schema
        self._statement = statement

    @property
    def routing_group(self):
        """
        The query string.
        :rtype: str
        """
        return self._routing_group

    @property
    def catalog(self):
        """
        :rtype: int
        """
        return self._catalog

    @property
    def schema(self):
        """
        :rtype: int
        """
        return self._schema

    @property
    def statement(self):
        """
        :rtype: int
        """
        return self._statement

    def to_flyte_idl(self):
        """
        :rtype: _presto.PrestoQuery
        """
        return _presto.PrestoQuery(
            routing_group=self._routing_group,
            catalog=self._catalog,
            schema=self._schema,
            statement=self._statement,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param _presto.PrestoQuery pb2_object:
        :return: PrestoQuery
        """
        return cls(
            routing_group=pb2_object.routing_group,
            catalog=pb2_object.catalog,
            schema=pb2_object.schema,
            statement=pb2_object.statement,
        )



import logging as _logging

logger = _logging.getLogger("flytekit")

# create console handler and set level to debug
ch = _logging.StreamHandler()
ch.setLevel(_logging.DEBUG)

# create formatter
formatter = _logging.Formatter("%(asctime)s-%(name)s-%(levelname)s$ %(message)s")

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)



import os as _os
import sys as _sys
import uuid as _uuid

from flytekit.common.exceptions.user import FlyteUserException as _FlyteUserException
from flytekit.configuration import gcp as _gcp_config
from flytekit.interfaces import random as _flyte_random
from flytekit.interfaces.data import common as _common_data
from flytekit.tools import subprocess as _subprocess

if _sys.version_info >= (3,):
    from shutil import which as _which
else:
    from distutils.spawn import find_executable as _which


def _update_cmd_config_and_execute(cmd):
    env = _os.environ.copy()
    return _subprocess.check_call(cmd, env=env)


def _amend_path(path):
    return _os.path.join(path, "*") if not path.endswith("*") else path


class GCSProxy(_common_data.DataProxy):
    _GS_UTIL_CLI = "gsutil"

    def __init__(self, raw_output_data_prefix_override: str = None):
        """
        :param raw_output_data_prefix_override: Instead of relying on the AWS or GCS configuration (see
            S3_SHARD_FORMATTER for AWS and GCS_PREFIX for GCP) setting when computing the shard
            path (_get_shard_path), use this prefix instead as a base. This code assumes that the
            path passed in is correct. That is, an S3 path won't be passed in when running on GCP.
        """
        self._raw_output_data_prefix_override = raw_output_data_prefix_override

    @property
    def raw_output_data_prefix_override(self) -> str:
        return self._raw_output_data_prefix_override

    @staticmethod
    def _check_binary():
        """
        Make sure that the `gsutil` cli is present
        """
        if not _which(GCSProxy._GS_UTIL_CLI):
            raise _FlyteUserException("gsutil (gcloud cli) not found! Please install.")

    @staticmethod
    def _maybe_with_gsutil_parallelism(*gsutil_args):
        """
        Check if we should run `gsutil` with the `-m` flag that enables
        parallelism via multiple threads/processes. Additional tweaking of
        this behavior can be achieved via the .boto configuration file. See:
        https://cloud.google.com/storage/docs/boto-gsutil
        """
        cmd = [GCSProxy._GS_UTIL_CLI]
        if _gcp_config.GSUTIL_PARALLELISM.get():
            cmd.append("-m")
        cmd.extend(gsutil_args)

        return cmd

    def exists(self, remote_path):
        """
        :param Text remote_path: remote gs:// path
        :rtype bool: whether the gs file exists or not
        """
        GCSProxy._check_binary()

        if not remote_path.startswith("gs://"):
            raise ValueError("Not an GS Key. Please use FQN (GS ARN) of the format gs://...")

        cmd = [GCSProxy._GS_UTIL_CLI, "-q", "stat", remote_path]
        try:
            _update_cmd_config_and_execute(cmd)
            return True
        except Exception:
            return False

    def download_directory(self, remote_path, local_path):
        """
        :param Text remote_path: remote gs:// path
        :param Text local_path: directory to copy to
        """
        GCSProxy._check_binary()

        if not remote_path.startswith("gs://"):
            raise ValueError("Not an GS Key. Please use FQN (GS ARN) of the format gs://...")

        cmd = self._maybe_with_gsutil_parallelism("cp", "-r", _amend_path(remote_path), local_path)
        return _update_cmd_config_and_execute(cmd)

    def download(self, remote_path, local_path):
        """
        :param Text remote_path: remote gs:// path
        :param Text local_path: directory to copy to
        """
        if not remote_path.startswith("gs://"):
            raise ValueError("Not an GS Key. Please use FQN (GS ARN) of the format gs://...")

        GCSProxy._check_binary()

        cmd = self._maybe_with_gsutil_parallelism("cp", remote_path, local_path)
        return _update_cmd_config_and_execute(cmd)

    def upload(self, file_path, to_path):
        """
        :param Text file_path:
        :param Text to_path:
        """
        GCSProxy._check_binary()

        cmd = self._maybe_with_gsutil_parallelism("cp", file_path, to_path)
        return _update_cmd_config_and_execute(cmd)

    def upload_directory(self, local_path, remote_path):
        """
        :param Text local_path:
        :param Text remote_path:
        """
        if not remote_path.startswith("gs://"):
            raise ValueError("Not an GS Key. Please use FQN (GS ARN) of the format gs://...")

        GCSProxy._check_binary()

        cmd = self._maybe_with_gsutil_parallelism(
            "cp",
            "-r",
            _amend_path(local_path),
            remote_path if remote_path.endswith("/") else remote_path + "/",
        )
        return _update_cmd_config_and_execute(cmd)

    def get_random_path(self) -> str:
        """
        If this object was created with a raw output data prefix, usually set by Propeller/Plugins at execution time
        and piped all the way here, it will be used instead of referencing the GCS_PREFIX configuration.
        """
        key = _uuid.UUID(int=_flyte_random.random.getrandbits(128)).hex
        prefix = self.raw_output_data_prefix_override or _gcp_config.GCS_PREFIX.get()
        return _os.path.join(prefix, key)

    def get_random_directory(self):
        """
        :rtype: Text
        """
        return self.get_random_path() + "/"



from typing import Dict, List, Optional

from flytekit.common import constants as _constants
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.mixins import hash as _hash_mixin
from flytekit.control_plane import identifier as _identifier
from flytekit.control_plane import interface as _interfaces
from flytekit.control_plane import nodes as _nodes
from flytekit.engines.flyte import engine as _flyte_engine
from flytekit.models import task as _task_models
from flytekit.models.core import identifier as _identifier_model
from flytekit.models.core import workflow as _workflow_models


class FlyteWorkflow(_hash_mixin.HashOnReferenceMixin, _workflow_models.WorkflowTemplate):
    """A Flyte control plane construct."""

    def __init__(
        self,
        nodes: List[_nodes.FlyteNode],
        interface,
        output_bindings,
        id,
        metadata,
        metadata_defaults,
    ):
        for node in nodes:
            for upstream in node.upstream_nodes:
                if upstream.id is None:
                    raise _user_exceptions.FlyteAssertion(
                        "Some nodes contained in the workflow were not found in the workflow description.  Please "
                        "ensure all nodes are either assigned to attributes within the class or an element in a "
                        "list, dict, or tuple which is stored as an attribute in the class."
                    )
        super(FlyteWorkflow, self).__init__(
            id=id,
            metadata=metadata,
            metadata_defaults=metadata_defaults,
            interface=interface,
            nodes=nodes,
            outputs=output_bindings,
        )
        self._flyte_nodes = nodes

    @property
    def upstream_entities(self):
        return set(n.executable_flyte_object for n in self._flyte_nodes)

    @property
    def interface(self) -> _interfaces.TypedInterface:
        return super(FlyteWorkflow, self).interface

    @property
    def entity_type_text(self) -> str:
        return "Workflow"

    @property
    def resource_type(self):
        return _identifier_model.ResourceType.WORKFLOW

    def get_sub_workflows(self) -> List["FlyteWorkflow"]:
        result = []
        for node in self.nodes:
            if node.workflow_node is not None and node.workflow_node.sub_workflow_ref is not None:
                if (
                    node.executable_flyte_object is not None
                    and node.executable_flyte_object.entity_type_text == "Workflow"
                ):
                    result.append(node.executable_flyte_object)
                    result.extend(node.executable_flyte_object.get_sub_workflows())
                else:
                    raise _system_exceptions.FlyteSystemException(
                        "workflow node with subworkflow found but bad executable "
                        "object {}".format(node.executable_flyte_object)
                    )

            # get subworkflows in conditional branches
            if node.branch_node is not None:
                if_else: _workflow_models.IfElseBlock = node.branch_node.if_else
                leaf_nodes: List[_nodes.FlyteNode] = filter(
                    None,
                    [
                        if_else.case.then_node,
                        *([] if if_else.other is None else [x.then_node for x in if_else.other]),
                        if_else.else_node,
                    ],
                )
                for leaf_node in leaf_nodes:
                    exec_flyte_obj = leaf_node.executable_flyte_object
                    if exec_flyte_obj is not None and exec_flyte_obj.entity_type_text == "Workflow":
                        result.append(exec_flyte_obj)
                        result.extend(exec_flyte_obj.get_sub_workflows())

        return result

    @classmethod
    @_exception_scopes.system_entry_point
    def fetch(cls, project: str, domain: str, name: str, version: str):
        workflow_id = _identifier.Identifier(_identifier_model.ResourceType.WORKFLOW, project, domain, name, version)
        admin_workflow = _flyte_engine.get_client().get_workflow(workflow_id)
        cwc = admin_workflow.closure.compiled_workflow
        flyte_workflow = cls.promote_from_model(
            base_model=cwc.primary.template,
            sub_workflows={sw.template.id: sw.template for sw in cwc.sub_workflows},
            tasks={t.template.id: t.template for t in cwc.tasks},
        )
        flyte_workflow._id = workflow_id
        return flyte_workflow

    @classmethod
    def get_non_system_nodes(cls, nodes: List[_workflow_models.Node]) -> List[_workflow_models.Node]:
        return [n for n in nodes if n.id not in {_constants.START_NODE_ID, _constants.END_NODE_ID}]

    @classmethod
    def promote_from_model(
        cls,
        base_model: _workflow_models.WorkflowTemplate,
        sub_workflows: Optional[Dict[_identifier.Identifier, _workflow_models.WorkflowTemplate]] = None,
        tasks: Optional[Dict[_identifier.Identifier, _task_models.TaskTemplate]] = None,
    ) -> "FlyteWorkflow":
        base_model_non_system_nodes = cls.get_non_system_nodes(base_model.nodes)
        sub_workflows = sub_workflows or {}
        tasks = tasks or {}
        node_map = {
            n.id: _nodes.FlyteNode.promote_from_model(n, sub_workflows, tasks) for n in base_model_non_system_nodes
        }

        # Set upstream nodes for each node
        for n in base_model_non_system_nodes:
            current = node_map[n.id]
            for upstream_id in current.upstream_node_ids:
                upstream_node = node_map[upstream_id]
                current << upstream_node

        # No inputs/outputs specified, see the constructor for more information on the overrides.
        return cls(
            nodes=list(node_map.values()),
            id=_identifier.Identifier.promote_from_model(base_model.id),
            metadata=base_model.metadata,
            metadata_defaults=base_model.metadata_defaults,
            interface=_interfaces.TypedInterface.promote_from_model(base_model.interface),
            output_bindings=base_model.outputs,
        )

    @_exception_scopes.system_entry_point
    def register(self, project, domain, name, version):
        # TODO
        pass

    @_exception_scopes.system_entry_point
    def serialize(self):
        # TODO
        pass

    @_exception_scopes.system_entry_point
    def validate(self):
        # TODO
        pass

    @_exception_scopes.system_entry_point
    def create_launch_plan(self, *args, **kwargs):
        # TODO
        pass

    @_exception_scopes.system_entry_point
    def __call__(self, *args, **input_map):
        raise NotImplementedError



from flytekit.common import constants as _common_constants
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.tasks import sdk_runnable as _sdk_runnable
from flytekit.contrib.sensors.base_sensor import Sensor as _Sensor


class SensorTask(_sdk_runnable.SdkRunnableTask):
    def _execute_user_code(self, context, inputs):
        sensor = super(SensorTask, self)._execute_user_code(context=context, inputs=inputs)
        if sensor is not None:
            if not isinstance(sensor, _Sensor):
                raise _user_exceptions.FlyteTypeException(
                    received_type=type(sensor),
                    expected_type=_Sensor,
                )
            succeeded = sensor.sense()
            if not succeeded:
                raise _user_exceptions.FlyteRecoverableException()


def sensor_task(
    _task_function=None,
    retries=0,
    interruptible=None,
    deprecated="",
    storage_request=None,
    cpu_request=None,
    gpu_request=None,
    memory_request=None,
    storage_limit=None,
    cpu_limit=None,
    gpu_limit=None,
    memory_limit=None,
    timeout=None,
    environment=None,
    cls=None,
):
    """
    Decorator to create a Sensor Task definition.  This task will run as a single unit of work on the platform.

    .. code-block:: python
        @sensor_task(retries=3)
        def my_task(wf_params):
            return HiveTableSensor(
                schema='default',
                table_name='mocked_table',
                host='localhost',
                port=1234,
            )

    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs.  No keyword
        arguments are allowed for wrapped task functions.
    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.
        .. note::
            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.
    :param bool interruptible: Specify whether task is interruptible
    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.
    :param Text storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for the task to run.  Default is set by platform-level configuration.
        .. note::
            This is currently not supported by the platform.
    :param Text cpu_request: [optional] Kubernetes resource string for lower-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs.
        Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for the task to execute.  Default is set by platform-level configuration.
        TODO: Add links to resource string documentation for Kubernetes
    :param Text storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for the task to run.  This amount is not guaranteed!  If not specified, it is set equal to storage_request.
        .. note::
            This is currently not supported by the platform.
    :param Text cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.
    :param Text gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs. This amount is not
        guaranteed!  If not specified, it is set equal to gpu_request.
    :param Text memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for the task to execute.  This amount is not guaranteed!  If not specified, it is set equal to
        memory_request.
    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.
    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.
    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  A user can use this to
        inject bespoke logic into the base Flyte programming model. Ideally, should be a sub-class of SensorTask or
        otherwise mimic the behavior.
    :rtype: SensorTask
    """

    def wrapper(fn):
        return (SensorTask or cls)(
            task_function=fn,
            task_type=_common_constants.SdkTaskType.SENSOR_TASK,
            retries=retries,
            interruptible=interruptible,
            deprecated=deprecated,
            storage_request=storage_request,
            cpu_request=cpu_request,
            gpu_request=gpu_request,
            memory_request=memory_request,
            storage_limit=storage_limit,
            cpu_limit=cpu_limit,
            gpu_limit=gpu_limit,
            memory_limit=memory_limit,
            timeout=timeout,
            environment=environment,
            custom={},
            discovery_version="",
            discoverable=False,
        )

    # This is syntactic-sugar, so that when calling this decorator without args, you can either
    # do it with () or without any ()
    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper



import json as _json
import typing

import six as _six
from flyteidl.admin import task_pb2 as _admin_task
from flyteidl.core import compiler_pb2 as _compiler
from flyteidl.core import literals_pb2 as _literals_pb2
from flyteidl.core import tasks_pb2 as _core_task
from flyteidl.plugins import pytorch_pb2 as _pytorch_task
from flyteidl.plugins import spark_pb2 as _spark_task
from flyteidl.plugins import tensorflow_pb2 as _tensorflow_task
from google.protobuf import json_format as _json_format
from google.protobuf import struct_pb2 as _struct

from flytekit.common.exceptions import user as _user_exceptions
from flytekit.models import common as _common
from flytekit.models import interface as _interface
from flytekit.models import literals as _literals
from flytekit.models import security as _sec
from flytekit.models.core import identifier as _identifier
from flytekit.plugins import flyteidl as _lazy_flyteidl
from flytekit.sdk.spark_types import SparkType as _spark_type


class Resources(_common.FlyteIdlEntity):
    class ResourceName(object):
        UNKNOWN = _core_task.Resources.UNKNOWN
        CPU = _core_task.Resources.CPU
        GPU = _core_task.Resources.GPU
        MEMORY = _core_task.Resources.MEMORY
        STORAGE = _core_task.Resources.STORAGE

    class ResourceEntry(_common.FlyteIdlEntity):
        def __init__(self, name, value):
            """
            :param int name: enum value from ResourceName
            :param Text value: a textual value describing the resource need.  Must be a valid k8s quantity.
            """
            self._name = name
            self._value = value

        @property
        def name(self):
            """
            enum value from ResourceName
            :rtype: int
            """
            return self._name

        @property
        def value(self):
            """
            A textual value describing the resource need.  Must be a valid k8s quantity.
            :rtype: Text
            """
            return self._value

        def to_flyte_idl(self):
            """
            :rtype: flyteidl.core.tasks_pb2.ResourceEntry
            """
            return _core_task.Resources.ResourceEntry(name=self.name, value=self.value)

        @classmethod
        def from_flyte_idl(cls, pb2_object):
            """
            :param flyteidl.core.tasks_pb2.Resources.ResourceEntry pb2_object:
            :rtype: Resources.ResourceEntry
            """
            return cls(name=pb2_object.name, value=pb2_object.value)

    def __init__(self, requests, limits):
        """
        :param list[Resources.ResourceEntry] requests: The desired resources for execution.  This is given on a best
            effort basis.
        :param list[Resources.ResourceEntry] limits: These are the limits required.  These are guaranteed to be
            satisfied.
        """
        self._requests = requests
        self._limits = limits

    @property
    def requests(self):
        """
        The desired resources for execution.  This is given on a best effort basis.
        :rtype: list[Resources.ResourceEntry]
        """
        return self._requests

    @property
    def limits(self):
        """
        These are the limits required.  These are guaranteed to be satisfied.
        :rtype: list[Resources.ResourceEntry]
        """
        return self._limits

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.tasks_pb2.Resources
        """
        return _core_task.Resources(
            requests=[r.to_flyte_idl() for r in self.requests],
            limits=[r.to_flyte_idl() for r in self.limits],
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.tasks_pb2.Resources.ResourceEntry pb2_object:
        :rtype: Resources
        """
        return cls(
            requests=[Resources.ResourceEntry.from_flyte_idl(r) for r in pb2_object.requests],
            limits=[Resources.ResourceEntry.from_flyte_idl(l) for l in pb2_object.limits],
        )


class RuntimeMetadata(_common.FlyteIdlEntity):
    class RuntimeType(object):
        OTHER = 0
        FLYTE_SDK = 1

    def __init__(self, type, version, flavor):
        """
        :param int type: Enum type from RuntimeMetadata.RuntimeType
        :param Text version: Version string for SDK version.  Can be used for metrics or managing breaking changes in
            Admin or Propeller
        :param Text flavor: Optional extra information about runtime environment (e.g. Python, GoLang, etc.)
        """
        self._type = type
        self._version = version
        self._flavor = flavor

    @property
    def type(self):
        """
        Enum type from RuntimeMetadata.RuntimeType
        :rtype: int
        """
        return self._type

    @property
    def version(self):
        """
        Version string for SDK version.  Can be used for metrics or managing breaking changes in Admin or Propeller
        :rtype: Text
        """
        return self._version

    @property
    def flavor(self):
        """
        Optional extra information about runtime environment (e.g. Python, GoLang, etc.)
        :rtype: Text
        """
        return self._flavor

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.tasks_pb2.RuntimeMetadata
        """
        return _core_task.RuntimeMetadata(type=self.type, version=self.version, flavor=self.flavor)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.tasks_pb2.RuntimeMetadata pb2_object:
        :rtype: RuntimeMetadata
        """
        return cls(type=pb2_object.type, version=pb2_object.version, flavor=pb2_object.flavor)


class TaskMetadata(_common.FlyteIdlEntity):
    def __init__(
        self,
        discoverable,
        runtime,
        timeout,
        retries,
        interruptible,
        discovery_version,
        deprecated_error_message,
    ):
        """
        Information needed at runtime to determine behavior such as whether or not outputs are discoverable, timeouts,
        and retries.

        :param bool discoverable: Whether or not the outputs of this task should be cached for discovery.
        :param RuntimeMetadata runtime: Metadata describing the runtime environment for this task.
        :param datetime.timedelta timeout: The amount of time to wait before timing out.  This includes queuing and
            scheduler latency.
        :param bool interruptible: Whether or not the task is interruptible.
        :param flytekit.models.literals.RetryStrategy retries: Retry strategy for this task.  0 retries means only
            try once.
        :param Text discovery_version: This is the version used to create a logical version for data in the cache.
            This is only used when `discoverable` is true.  Data is considered discoverable if: the inputs to a given
            task are the same and the discovery_version is also the same.
        :param Text deprecated: This string can be used to mark the task as deprecated.  Consumers of the task will
            receive deprecation warnings.
        """
        self._discoverable = discoverable
        self._runtime = runtime
        self._timeout = timeout
        self._interruptible = interruptible
        self._retries = retries
        self._discovery_version = discovery_version
        self._deprecated_error_message = deprecated_error_message

    @property
    def discoverable(self):
        """
        Whether or not the outputs of this task should be cached for discovery.
        :rtype: bool
        """
        return self._discoverable

    @property
    def runtime(self):
        """
        Metadata describing the runtime environment for this task.
        :rtype: RuntimeMetadata
        """
        return self._runtime

    @property
    def retries(self):
        """
        Retry strategy for this task.  0 retries means only try once.
        :rtype: flytekit.models.literals.RetryStrategy
        """
        return self._retries

    @property
    def timeout(self):
        """
        The amount of time to wait before timing out.  This includes queuing and scheduler latency.
        :rtype: datetime.timedelta
        """
        return self._timeout

    @property
    def interruptible(self):
        """
        Whether or not the task is interruptible.
        :rtype: bool
        """
        return self._interruptible

    @property
    def discovery_version(self):
        """
        This is the version used to create a logical version for data in the cache.
        This is only used when `discoverable` is true.  Data is considered discoverable if: the inputs to a given
        task are the same and the discovery_version is also the same.
        :rtype: Text
        """
        return self._discovery_version

    @property
    def deprecated_error_message(self):
        """
        This string can be used to mark the task as deprecated.  Consumers of the task will receive deprecation
        warnings.
        :rtype: Text
        """
        return self._deprecated_error_message

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.task_pb2.TaskMetadata
        """
        tm = _core_task.TaskMetadata(
            discoverable=self.discoverable,
            runtime=self.runtime.to_flyte_idl(),
            retries=self.retries.to_flyte_idl(),
            interruptible=self.interruptible,
            discovery_version=self.discovery_version,
            deprecated_error_message=self.deprecated_error_message,
        )
        if self.timeout:
            tm.timeout.FromTimedelta(self.timeout)
        return tm

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.task_pb2.TaskMetadata pb2_object:
        :rtype: TaskMetadata
        """
        return cls(
            discoverable=pb2_object.discoverable,
            runtime=RuntimeMetadata.from_flyte_idl(pb2_object.runtime),
            timeout=pb2_object.timeout.ToTimedelta(),
            interruptible=pb2_object.interruptible if pb2_object.HasField("interruptible") else None,
            retries=_literals.RetryStrategy.from_flyte_idl(pb2_object.retries),
            discovery_version=pb2_object.discovery_version,
            deprecated_error_message=pb2_object.deprecated_error_message,
        )


class TaskTemplate(_common.FlyteIdlEntity):
    def __init__(
        self,
        id,
        type,
        metadata,
        interface,
        custom,
        container=None,
        task_type_version=0,
        security_context=None,
        config=None,
    ):
        """
        A task template represents the full set of information necessary to perform a unit of work in the Flyte system.
        It contains the metadata about what inputs and outputs are consumed or produced.  It also contains the metadata
        necessary for Flyte Propeller to do the appropriate work.

        :param flytekit.models.core.identifier.Identifier id: This is generated by the system and uniquely identifies
            the task.
        :param Text type: This is used to define additional extensions for use by Propeller or SDK.
        :param TaskMetadata metadata: This contains information needed at runtime to determine behavior such as
            whether or not outputs are discoverable, timeouts, and retries.
        :param flytekit.models.interface.TypedInterface interface: The interface definition for this task.
        :param dict[Text, T] custom: Dictionary that must be serializable to a protobuf Struct for custom task plugins.
        :param Container container: Provides the necessary entrypoint information for execution.  For instance,
            a Container might be specified with the necessary command line arguments.
        :param int task_type_version: Specific version of this task type used by plugins to potentially modify
            execution behavior or serialization.
        :param dict[str, str] config: For plugin tasks this represents additional configuration information to be used
            in tandem with the custom.
        :param dict[str, str] config: For plugin tasks this represents additional configuration information to be used
            in tandem with the custom.
        """
        self._id = id
        self._type = type
        self._metadata = metadata
        self._interface = interface
        self._custom = custom
        self._container = container
        self._task_type_version = task_type_version
        self._config = config
        self._security_context = security_context

    @property
    def id(self):
        """
        This is generated by the system and uniquely identifies the task.
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._id

    @property
    def type(self):
        """
        This is used to identify additional extensions for use by Propeller or SDK.
        :rtype: Text
        """
        return self._type

    @property
    def metadata(self):
        """
        This contains information needed at runtime to determine behavior such as whether or not outputs are
        discoverable, timeouts, and retries.
        :rtype: TaskMetadata
        """
        return self._metadata

    @property
    def interface(self):
        """
        The interface definition for this task.
        :rtype: flytekit.models.interface.TypedInterface
        """
        return self._interface

    @property
    def custom(self):
        """
        Arbitrary dictionary containing metadata for custom plugins.
        :rtype: dict[Text, T]
        """
        return self._custom

    @property
    def task_type_version(self):
        return self._task_type_version

    @property
    def container(self):
        """
        If not None, the target of execution should be a container.
        :rtype: Container
        """
        return self._container

    @property
    def config(self):
        """
        Arbitrary dictionary containing metadata for parsing and handling custom plugins.
        :rtype: dict[Text, T]
        """
        return self._config

    @property
    def security_context(self):
        return self._security_context

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.tasks_pb2.TaskTemplate
        """
        task_template = _core_task.TaskTemplate(
            id=self.id.to_flyte_idl(),
            type=self.type,
            metadata=self.metadata.to_flyte_idl(),
            interface=self.interface.to_flyte_idl(),
            custom=_json_format.Parse(_json.dumps(self.custom), _struct.Struct()) if self.custom else None,
            container=self.container.to_flyte_idl() if self.container else None,
            task_type_version=self.task_type_version,
            security_context=self.security_context.to_flyte_idl() if self.security_context else None,
            config={k: v for k, v in self.config.items()} if self.config is not None else None,
        )
        return task_template

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.tasks_pb2.TaskTemplate pb2_object:
        :rtype: TaskTemplate
        """
        return cls(
            id=_identifier.Identifier.from_flyte_idl(pb2_object.id),
            type=pb2_object.type,
            metadata=TaskMetadata.from_flyte_idl(pb2_object.metadata),
            interface=_interface.TypedInterface.from_flyte_idl(pb2_object.interface),
            custom=_json_format.MessageToDict(pb2_object.custom) if pb2_object else None,
            container=Container.from_flyte_idl(pb2_object.container) if pb2_object.HasField("container") else None,
            task_type_version=pb2_object.task_type_version,
            security_context=_sec.SecurityContext.from_flyte_idl(pb2_object.security_context)
            if pb2_object.security_context
            else None,
            config={k: v for k, v in pb2_object.config.items()} if pb2_object.config is not None else None,
        )


class TaskSpec(_common.FlyteIdlEntity):
    def __init__(self, template):
        """
        :param TaskTemplate template:
        """
        self._template = template

    @property
    def template(self):
        """
        :rtype: TaskTemplate
        """
        return self._template

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.tasks_pb2.TaskSpec
        """
        return _admin_task.TaskSpec(template=self.template.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.tasks_pb2.TaskSpec pb2_object:
        :rtype: TaskSpec
        """
        return cls(TaskTemplate.from_flyte_idl(pb2_object.template))


class Task(_common.FlyteIdlEntity):
    def __init__(self, id, closure):
        """
        :param flytekit.models.core.identifier.Identifier id: The (project, domain, name) identifier for this task.
        :param TaskClosure closure: The closure for the underlying workload.
        """
        self._id = id
        self._closure = closure

    @property
    def id(self):
        """
        The (project, domain, name, version) identifier for this task.
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._id

    @property
    def closure(self):
        """
        The closure for the underlying workload.
        :rtype: TaskClosure
        """
        return self._closure

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.task_pb2.Task
        """
        return _admin_task.Task(
            closure=self.closure.to_flyte_idl(),
            id=self.id.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.task_pb2.Task pb2_object:
        :rtype: TaskDefinition
        """
        return cls(
            closure=TaskClosure.from_flyte_idl(pb2_object.closure),
            id=_identifier.Identifier.from_flyte_idl(pb2_object.id),
        )


class TaskClosure(_common.FlyteIdlEntity):
    def __init__(self, compiled_task):
        """
        :param CompiledTask compiled_task:
        """
        self._compiled_task = compiled_task

    @property
    def compiled_task(self):
        """
        :rtype: CompiledTask
        """
        return self._compiled_task

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.task_pb2.TaskClosure
        """
        return _admin_task.TaskClosure(compiled_task=self.compiled_task.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.task_pb2.TaskClosure pb2_object:
        :rtype: TaskClosure
        """
        return cls(compiled_task=CompiledTask.from_flyte_idl(pb2_object.compiled_task))


class CompiledTask(_common.FlyteIdlEntity):
    def __init__(self, template):
        """
        :param TaskTemplate template:
        """
        self._template = template

    @property
    def template(self):
        """
        :rtype: TaskTemplate
        """
        return self._template

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.compiler_pb2.CompiledTask
        """
        return _compiler.CompiledTask(template=self.template.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.compiler_pb2.CompiledTask pb2_object:
        :rtype: CompiledTask
        """
        return cls(template=TaskTemplate.from_flyte_idl(pb2_object.template))


class SparkJob(_common.FlyteIdlEntity):
    def __init__(
        self,
        spark_type,
        application_file,
        main_class,
        spark_conf,
        hadoop_conf,
        executor_path,
    ):
        """
        This defines a SparkJob target.  It will execute the appropriate SparkJob.

        :param application_file: The main application file to execute.
        :param dict[Text, Text] spark_conf: A definition of key-value pairs for spark config for the job.
        :param dict[Text, Text] hadoop_conf: A definition of key-value pairs for hadoop config for the job.
        """
        self._application_file = application_file
        self._spark_type = spark_type
        self._main_class = main_class
        self._executor_path = executor_path
        self._spark_conf = spark_conf
        self._hadoop_conf = hadoop_conf

    def with_overrides(
        self, new_spark_conf: typing.Dict[str, str] = None, new_hadoop_conf: typing.Dict[str, str] = None
    ) -> "SparkJob":
        if not new_spark_conf:
            new_spark_conf = self.spark_conf

        if not new_hadoop_conf:
            new_hadoop_conf = self.hadoop_conf

        return SparkJob(
            spark_type=self.spark_type,
            application_file=self.application_file,
            main_class=self.main_class,
            spark_conf=new_spark_conf,
            hadoop_conf=new_hadoop_conf,
            executor_path=self.executor_path,
        )

    @property
    def main_class(self):
        """
        The main class to execute
        :rtype: Text
        """
        return self._main_class

    @property
    def spark_type(self):
        """
        Spark Job Type
        :rtype: Text
        """
        return self._spark_type

    @property
    def application_file(self):
        """
        The main application file to execute
        :rtype: Text
        """
        return self._application_file

    @property
    def executor_path(self):
        """
        The python executable to use
        :rtype: Text
        """
        return self._executor_path

    @property
    def spark_conf(self):
        """
        A definition of key-value pairs for spark config for the job.
         :rtype: dict[Text, Text]
        """
        return self._spark_conf

    @property
    def hadoop_conf(self):
        """
         A definition of key-value pairs for hadoop config for the job.
        :rtype: dict[Text, Text]
        """
        return self._hadoop_conf

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.plugins.spark_pb2.SparkJob
        """

        if self.spark_type == _spark_type.PYTHON:
            application_type = _spark_task.SparkApplication.PYTHON
        elif self.spark_type == _spark_type.JAVA:
            application_type = _spark_task.SparkApplication.JAVA
        elif self.spark_type == _spark_type.SCALA:
            application_type = _spark_task.SparkApplication.SCALA
        elif self.spark_type == _spark_type.R:
            application_type = _spark_task.SparkApplication.R
        else:
            raise _user_exceptions.FlyteValidationException("Invalid Spark Application Type Specified")

        return _spark_task.SparkJob(
            applicationType=application_type,
            mainApplicationFile=self.application_file,
            mainClass=self.main_class,
            executorPath=self.executor_path,
            sparkConf=self.spark_conf,
            hadoopConf=self.hadoop_conf,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.plugins.spark_pb2.SparkJob pb2_object:
        :rtype: SparkJob
        """

        application_type = _spark_type.PYTHON
        if pb2_object.type == _spark_task.SparkApplication.JAVA:
            application_type = _spark_type.JAVA
        elif pb2_object.type == _spark_task.SparkApplication.SCALA:
            application_type = _spark_type.SCALA
        elif pb2_object.type == _spark_task.SparkApplication.R:
            application_type = _spark_type.R

        return cls(
            type=application_type,
            spark_conf=pb2_object.sparkConf,
            application_file=pb2_object.mainApplicationFile,
            main_class=pb2_object.mainClass,
            hadoop_conf=pb2_object.hadoopConf,
            executor_path=pb2_object.executorPath,
        )


class IOStrategy(_common.FlyteIdlEntity):
    """
    Provides methods to manage data in and out of the Raw container using Download Modes. This can only be used if DataLoadingConfig is enabled.
    """

    DOWNLOAD_MODE_EAGER = _core_task.IOStrategy.DOWNLOAD_EAGER
    DOWNLOAD_MODE_STREAM = _core_task.IOStrategy.DOWNLOAD_STREAM
    DOWNLOAD_MODE_NO_DOWNLOAD = _core_task.IOStrategy.DO_NOT_DOWNLOAD

    UPLOAD_MODE_EAGER = _core_task.IOStrategy.UPLOAD_EAGER
    UPLOAD_MODE_ON_EXIT = _core_task.IOStrategy.UPLOAD_ON_EXIT
    UPLOAD_MODE_NO_UPLOAD = _core_task.IOStrategy.DO_NOT_UPLOAD

    def __init__(
        self,
        download_mode: _core_task.IOStrategy.DownloadMode = DOWNLOAD_MODE_EAGER,
        upload_mode: _core_task.IOStrategy.UploadMode = UPLOAD_MODE_ON_EXIT,
    ):
        self._download_mode = download_mode
        self._upload_mode = upload_mode

    def to_flyte_idl(self) -> _core_task.IOStrategy:
        return _core_task.IOStrategy(download_mode=self._download_mode, upload_mode=self._upload_mode)

    @classmethod
    def from_flyte_idl(cls, pb2_object: _core_task.IOStrategy):
        if pb2_object is None:
            return None
        return cls(
            download_mode=pb2_object.download_mode,
            upload_mode=pb2_object.upload_mode,
        )


class DataLoadingConfig(_common.FlyteIdlEntity):
    LITERALMAP_FORMAT_PROTO = _core_task.DataLoadingConfig.PROTO
    LITERALMAP_FORMAT_JSON = _core_task.DataLoadingConfig.JSON
    LITERALMAP_FORMAT_YAML = _core_task.DataLoadingConfig.YAML
    _LITERALMAP_FORMATS = frozenset([LITERALMAP_FORMAT_JSON, LITERALMAP_FORMAT_PROTO, LITERALMAP_FORMAT_YAML])

    def __init__(
        self,
        input_path: str,
        output_path: str,
        enabled: bool = True,
        format: _core_task.DataLoadingConfig.LiteralMapFormat = LITERALMAP_FORMAT_PROTO,
        io_strategy: IOStrategy = None,
    ):
        if format not in self._LITERALMAP_FORMATS:
            raise ValueError(
                "Metadata format {} not supported. Should be one of {}".format(format, self._LITERALMAP_FORMATS)
            )
        self._input_path = input_path
        self._output_path = output_path
        self._enabled = enabled
        self._format = format
        self._io_strategy = io_strategy

    def to_flyte_idl(self) -> _core_task.DataLoadingConfig:
        return _core_task.DataLoadingConfig(
            input_path=self._input_path,
            output_path=self._output_path,
            format=self._format,
            enabled=self._enabled,
            io_strategy=self._io_strategy.to_flyte_idl() if self._io_strategy is not None else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2: _core_task.DataLoadingConfig) -> "DataLoadingConfig":
        if pb2 is None:
            return None
        return cls(
            input_path=pb2.input_path,
            output_path=pb2.output_path,
            enabled=pb2.enabled,
            format=pb2.format,
            io_strategy=IOStrategy.from_flyte_idl(pb2.io_strategy) if pb2.HasField("io_strategy") else None,
        )


class Container(_common.FlyteIdlEntity):
    def __init__(self, image, command, args, resources, env, config, data_loading_config=None):
        """
        This defines a container target.  It will execute the appropriate command line on the appropriate image with
        the given configurations.

        :param Text image: The fully-qualified identifier for the image.
        :param list[Text] command: A list of 'words' for the command.  i.e. ['aws', 's3', 'ls']
        :param list[Text] args: A list of arguments for the command.  i.e. ['s3://some/path', '/tmp/local/path']
        :param Resources resources: A definition of requisite compute resources.
        :param dict[Text, Text] env: A definition of key-value pairs for environment variables.
        :param dict[Text, Text] config: A definition of configuration key-value pairs.
        :type DataLoadingConfig data_loading_config: object
        """
        self._data_loading_config = data_loading_config
        self._image = image
        self._command = command
        self._args = args
        self._resources = resources
        self._env = env
        self._config = config

    @property
    def image(self):
        """
        The fully-qualified identifier for the image.
        :rtype: Text
        """
        return self._image

    @property
    def command(self):
        """
        A list of 'words' for the command.  i.e. ['aws', 's3', 'ls']
        :rtype: list[Text]
        """
        return self._command

    @property
    def args(self):
        """
         A list of arguments for the command.  i.e. ['s3://some/path', '/tmp/local/path']
        :rtype: list[Text]
        """
        return self._args

    @property
    def resources(self):
        """
        A definition of requisite compute resources.
        :rtype: Resources
        """
        return self._resources

    @property
    def env(self):
        """
        A definition of key-value pairs for environment variables.  Currently, only str->str is
            supported.
        :rtype: dict[Text, Text]
        """
        return self._env

    @property
    def config(self):
        """
        A definition of key-value pairs for configuration.  Currently, only str->str is
            supported.
        :rtype: dict[Text, Text]
        """
        return self._config

    @property
    def data_loading_config(self):
        """
        :rtype: DataLoadingConfig
        """
        return self._data_loading_config

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.tasks_pb2.Container
        """
        return _core_task.Container(
            image=self.image,
            command=self.command,
            args=self.args,
            resources=self.resources.to_flyte_idl(),
            env=[_literals_pb2.KeyValuePair(key=k, value=v) for k, v in _six.iteritems(self.env)],
            config=[_literals_pb2.KeyValuePair(key=k, value=v) for k, v in _six.iteritems(self.config)],
            data_config=self._data_loading_config.to_flyte_idl() if self._data_loading_config else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.task_pb2.Task pb2_object:
        :rtype: Container
        """
        return cls(
            image=pb2_object.image,
            command=pb2_object.command,
            args=pb2_object.args,
            resources=Resources.from_flyte_idl(pb2_object.resources),
            env={kv.key: kv.value for kv in pb2_object.env},
            config={kv.key: kv.value for kv in pb2_object.config},
            data_loading_config=DataLoadingConfig.from_flyte_idl(pb2_object.data_config)
            if pb2_object.HasField("data_config")
            else None,
        )


class SidecarJob(_common.FlyteIdlEntity):
    def __init__(self, pod_spec, primary_container_name):
        """
        A sidecar job represents the full kubernetes pod spec and related metadata required for executing a sidecar
        task.

        :param pod_spec: k8s.io.api.core.v1.PodSpec
        :param primary_container_name: Text
        """
        self._pod_spec = pod_spec
        self._primary_container_name = primary_container_name

    @property
    def pod_spec(self):
        """
        :rtype: k8s.io.api.core.v1.PodSpec
        """
        return self._pod_spec

    @property
    def primary_container_name(self):
        """
        :rtype: Text
        """
        return self._primary_container_name

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.tasks_pb2.SidecarJob
        """
        return _lazy_flyteidl.plugins.sidecar_pb2.SidecarJob(
            pod_spec=self.pod_spec, primary_container_name=self.primary_container_name
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.task_pb2.Task pb2_object:
        :rtype: Container
        """
        return cls(
            pod_spec=pb2_object.pod_spec,
            primary_container_name=pb2_object.primary_container_name,
        )


class PyTorchJob(_common.FlyteIdlEntity):
    def __init__(self, workers_count):
        self._workers_count = workers_count

    @property
    def workers_count(self):
        return self._workers_count

    def to_flyte_idl(self):
        return _pytorch_task.DistributedPyTorchTrainingTask(
            workers=self.workers_count,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        return cls(
            workers_count=pb2_object.workers,
        )


class TensorFlowJob(_common.FlyteIdlEntity):
    def __init__(self, workers_count, ps_replicas_count, chief_replicas_count):
        self._workers_count = workers_count
        self._ps_replicas_count = ps_replicas_count
        self._chief_replicas_count = chief_replicas_count

    @property
    def workers_count(self):
        return self._workers_count

    @property
    def ps_replicas_count(self):
        return self._ps_replicas_count

    @property
    def chief_replicas_count(self):
        return self._chief_replicas_count

    def to_flyte_idl(self):
        return _tensorflow_task.DistributedTensorflowTrainingTask(
            workers=self.workers_count, ps_replicas=self.ps_replicas_count, chief_replicas=self.chief_replicas_count
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        return cls(
            workers_count=pb2_object.workers,
            ps_replicas_count=pb2_object.ps_replicas,
            chief_replicas_count=pb2_object.chief_replicas,
        )
